# -*- coding: utf-8 -*-

import numpy as np

from astropy.coordinates import (
    SkyCoord,
)


def rotate_spherical_vector(ra1, dec1, ra2, dec2, ra3, dec3):
    """Calculates the rotation matrix R to rotate the spherical vector
    (ra1,dec1) onto the direction (ra2,dec2), and performs this rotation on the
    spherical vector (ra3,dec3).

    In practice (ra1,dec1) refers to the true location of a MC event,
    (ra2,dec2) the true location of the signal source, and (ra3,dec3) the
    reconstructed location of the MC event, which should get rotated according
    to the rotation of the two true directions.
    """
    # Make sure, the inputs are 1D arrays.
    ra1 = np.atleast_1d(ra1)
    dec1 = np.atleast_1d(dec1)
    ra2 = np.atleast_1d(ra2)
    dec2 = np.atleast_1d(dec2)
    ra3 = np.atleast_1d(ra3)
    dec3 = np.atleast_1d(dec3)

    assert (
        len(ra1) == len(dec1) ==
        len(ra2) == len(dec2) ==
        len(ra3) == len(dec3)
    ), 'All input argument arrays must be of the same length!'

    N_event = len(ra1)

    # Calculate the space angle alpha between vector 1 and vector 2, and
    # correct for possible rounding errors.
    cos_alpha = (np.cos(ra2 - ra1) * np.cos(dec1) * np.cos(dec2)
                 + np.sin(dec1) * np.sin(dec2))
    cos_alpha[cos_alpha > 1] = 1
    cos_alpha[cos_alpha < -1] = -1
    alpha = np.arccos(cos_alpha)

    # Define the three 3D-vectors in spherical coordinate system. Each vector
    # is a (N_event,3)-shaped 2D array.
    vec1 = np.vstack([np.cos(ra1) * np.cos(dec1),
                      np.sin(ra1) * np.cos(dec1),
                      np.sin(dec1)]).T
    vec2 = np.vstack([np.cos(ra2) * np.cos(dec2),
                      np.sin(ra2) * np.cos(dec2),
                      np.sin(dec2)]).T
    vec3 = np.vstack([np.cos(ra3) * np.cos(dec3),
                      np.sin(ra3) * np.cos(dec3),
                      np.sin(dec3)]).T

    # Calculate the normalized rotation axis vector, nrot. nrot is a
    # (N_event,3)-shaped ndarray.
    nrot = np.cross(vec1, vec2)
    norm = np.sqrt(np.sum(nrot**2, axis=1))
    nrot[norm > 0] /= norm[np.newaxis, norm > 0].T

    # Define the diagonal 3D unit matrix.
    one = np.diagflat(np.ones(3))

    # Calculate the rotation matrix R_i for each event i and perform the
    # rotation on vector 3 for each event.
    vec = np.empty((N_event, 3), dtype=np.float64)

    sin_alpha = np.sin(alpha)
    twopi = 2*np.pi
    # Remap functions to avoid Python's (.)-resolution millions of times.
    (np_outer, np_dot, np_roll, np_diag, np_T) = (
     np.outer, np.dot, np.roll, np.diag, np.transpose)
    for i in range(N_event):
        cos_alpha_i = cos_alpha[i]
        nrot_i = nrot[i]
        nrotTnrot_i = np_outer(nrot_i, nrot_i)

        # Calculate cross product matrix, nrotx_i:
        # A[ij] = x_i * y_j - y_i * x_j
        skv = np_roll(np_roll(np_diag(nrot_i), shift=1, axis=1), shift=-1, axis=0)
        nrotx_i = skv - np_T(skv)

        # Calculate rotation matrix, R_i.
        R_i = ((1. - cos_alpha_i) * nrotTnrot_i
               + one*cos_alpha_i
               + sin_alpha[i] * nrotx_i)
        vec[i] = np_dot(R_i, np_T(vec3[i]))

    ra = np.arctan2(vec[:, 1], vec[:, 0])
    ra += np.where(ra < 0., twopi, 0.)
    dec = np.arcsin(vec[:, 2])

    return (ra, dec)


def rotate_signal_events_on_sphere(
        src_ra,
        src_dec,
        evt_true_ra,
        evt_true_dec,
        evt_reco_ra,
        evt_reco_dec,
):
    """Rotate signal events on a sphere to a given source position preserving
    position angle and separation (great circle distance) between the event's
    true and reco directions.

    Parameters
    ----------
    src_ra : instance of numpy.ndarray
        The (N_events,)-shaped 1D numpy.ndarray holding the true right-ascension
        of the source.
    src_dec : instance of numpy.ndarray
        The (N_events,)-shaped 1D numpy.ndarray holding the true declination
        of the source.
    evt_true_ra : instance of numpy.ndarray
        The (N_events,)-shaped 1D numpy.ndarray holding the true right-ascension
        of the MC event.
    evt_true_dec : instance of numpy.ndarray
        The (N_events,)-shaped 1D numpy.ndarray holding the true declination of
        the MC event.
    evt_reco_ra : instance of numpy.ndarray
        The (N_events,)-shaped 1D numpy.ndarray holding the reconstructed
        right-ascension of the MC event.
    evt_reco_dec : instance of numpy.ndarray
        The (N_events,)-shaped 1D numpy.ndarray holding the reconstructed
        declination of the MC event.

    Returns
    -------
    rot_evt_reco_ra : instance of numpy.ndarray
        The (N_events,)-shaped 1D numpy.ndarray holding the rotated
        reconstructed event right-ascension.
    rot_evt_reco_dec : instance of numpy.ndarray
        The (N_events,)-shaped 1D numpy.ndarray holding the rotated
        reconstructed event declination.
    """
    assert (
        len(src_ra) == len(src_dec) ==
        len(evt_true_ra) == len(evt_true_dec) ==
        len(evt_reco_ra) == len(evt_reco_dec)
    ), 'All input argument arrays must be of the same length!'

    v_source = SkyCoord(src_ra, src_dec, frame="icrs", unit="rad")
    v_evt_true = SkyCoord(evt_true_ra, evt_true_dec, frame="icrs", unit="rad")
    v_evt_reco = SkyCoord(evt_reco_ra, evt_reco_dec, frame="icrs", unit="rad")

    position_angle = v_evt_true.position_angle(v_evt_reco)
    separation = v_evt_true.separation(v_evt_reco)

    v_rotated = v_source.directional_offset_by(position_angle, separation)
    (rot_evt_reco_ra, rot_evt_reco_dec) = (v_rotated.ra.rad, v_rotated.dec.rad)

    return (rot_evt_reco_ra, rot_evt_reco_dec)


def angular_separation(ra1, dec1, ra2, dec2, psi_floor=None):
    """Calculates the angular separation on the sphere between two vectors on
    the sphere.

    Parameters
    ----------
    ra1 : instance of numpy.ndarray
        The (N_events,)-shaped numpy.ndarray holding the right-ascension or
        longitude coordinate of the first vector in radians.
    dec1 : instance of numpy.ndarray
        The (N_events,)-shaped numpy.ndarray holding declination or latitude
        coordinate of the first vector in radians.
    ra2 : instance of numpy.ndarray
        The (N_events,)-shaped numpy.ndarray holding the right-ascension or
        longitude coordinate of the second vector in radians.
    dec2 : instance of numpy.ndarray
        The (N_events,)-shaped numpy.ndarray holding declination coordinate of
        the second vector in radians.
    psi_floor : float | None
        If not ``None``, specifies the floor value of psi.

    Returns
    -------
    psi : instance of numpy.ndarray
        The (N_events,)-shaped numpy.ndarray holding the calculated angular
        separation value of each event.
    """
    delta_ra = np.abs(ra1 - ra2)
    delta_dec = np.abs(dec1 - dec2)

    x = np.sin(delta_dec / 2.)**2. +\
        np.cos(dec1) * np.cos(dec2) * np.sin(delta_ra / 2.)**2.

    # Handle possible floating precision errors.
    x[x < 0.] = 0.
    x[x > 1.] = 1.

    psi = 2. * np.arcsin(np.sqrt(x))

    if psi_floor is not None:
        psi = np.where(psi < psi_floor, psi_floor, psi)

    return psi



# -*- coding: utf-8 -*-

"""This module contains utility functions for the argument parser of an analysis
script.
"""

import argparse


def create_argparser(
        description=None,
        options=True):
    """Creates an argparser with the given description and adds common options
    useful for analysis scripts.

    Parameters
    ----------
    description : str | None
        The description for the argparser.
    options : bool | dict | None
        If set to None or False, no options will be added.
        If set to True, all common analysis script options will be added.
        If set to a dictionary, individual options can be turned on and off.
        See the :func:`add_argparser_options` for possible options.
        Default is ``True``.
    """
    parser = argparse.ArgumentParser(
        description=description,
        formatter_class=argparse.RawTextHelpFormatter
    )

    if options is True:
        options = dict()

    if isinstance(options, dict):
        add_argparser_options(
            parser=parser,
            **options)

    return parser


def add_argparser_options(
        parser,
        config=True,
        data_basepath=True,
        debug_logfile=True,
        enable_tracing=True,
        n_cpu=True,
        seed=True):
    """Adds common argparser options to the given argparser instance, useful for
    analysis scripts.

    Parameters
    ----------
    parser : instance of ArgumentParser
        The instance of ArgumentParser to which options should get added.
    config : bool
        If set to ``True``, the ``--config`` option of type ``str``
        will be added.
        It specifies the configuration file.
        The default value is ``None``.
        The option destination is ``config``.
    data_basepath : bool
        If set to ``True``, the ``--data-basepath`` option of type ``str``
        will be added.
        It specifies the base path to the data samples.
        The default value is ``None``.
        The option destination is ``data_basepath``.
    debug_logfile : bool
        If set to ``True``, the ``--debug-logfile`` option of type ``str``
        will be added.
        If not ``None``, it specifies the log file for dubug information.
        The default value is ``None``.
        The option destination is ``debug_logfile``.
    enable_tracing : bool
        If set to ``True``, the ``--enable-tracing`` option of type ``bool``
        will be added.
        If specified, enables the logging on the tracing level, i.e. a lot of
        DEBUG messages.
        The default value is ``False``.
        The option destination is ``enable_tracing``.
    n_cpu : bool
        If set to ``True``, the ``--n-cpu`` option of type ``int``
        will be added.
        It specifies the number of CPUs to utilize where parallelization is
        possible.
        The default value is ``1``.
        The option destination is ``n_cpu``.
    seed : bool
        If set to ``True``, the ``--seed`` option of type ``int``
        will be added.
        It specifies the seed for the random number generator.
        The default value is ``0``.
        The option destination is ``seed``.
    """
    if config:
        parser.add_argument(
            '--config',
            dest='config',
            default=None,
            type=str,
            help='The configuration file. '
                 '(default=None)')

    if data_basepath:
        parser.add_argument(
            '--data-basepath',
            dest='data_basepath',
            default=None,
            type=str,
            help='The base path to the data samples. '
                 '(default=None)')

    if debug_logfile:
        parser.add_argument(
            '--debug-logfile',
            dest='debug_logfile',
            default=None,
            type=str,
            help='If not None, it specifies the log file for dubug '
                 'information. '
                 '(default=None)')

    if enable_tracing:
        parser.add_argument(
            '--enable-tracing',
            dest='enable_tracing',
            default=False,
            action='store_true',
            help='If specified, enables the logging on the tracing level, i.e. '
                 'a lot of DEBUG messages. '
                 '(default=False)')

    if n_cpu:
        parser.add_argument(
            '--n-cpu',
            dest='n_cpu',
            default=1,
            type=int,
            help='The number of CPUs to utilize where parallelization is '
                 'possible. '
                 '(default=1)')

    if seed:
        parser.add_argument(
            '--seed',
            dest='seed',
            default=0,
            type=int,
            help='The seed for the random number generator. '
                 '(default=0)')



# -*- coding: utf-8 -*-

"""This module contains utility functions for logging functionalities of an
analysis script.
"""

import logging

from skyllh.core.debugging import (
    get_logger,
    setup_console_handler,
    setup_file_handler,
)


def setup_logging(
        cfg,
        script_logger_name,
        log_format=None,
        log_level=logging.INFO,
        debug_pathfilename=None):
    """Installs console handlers for the ``skyllh`` and ``script_logger_name``
    loggers. If a debug file is specified, file handlers for debug messages
    will be installed as well.

    Parameters
    ----------
    cfg : instance of Config
        The instance of Config holding the local configuration.
    script_logger_name : str
        The name of the logger used by the script.
    log_format : str | None
        The format template of the log message. If set to ``None``, the format
        will be taken from ``cfg['debugging']['log_format']``.
    log_level : int
        The log level of the loggers. The default is ``logging.INFO``.
    debug_pathfilename : str | None
        If not ``None``, file handlers for DEBUG messages will be installed and
        those messages will be stored in the given file.

    Returns
    -------
    script_logger : instance of logging.Logger
        The logger instance of the script, specified by ``script_logger_name``.
    """
    if log_format is None:
        log_format = cfg['debugging']['log_format']

    setup_console_handler(
        cfg=cfg,
        name='skyllh',
        log_level=log_level,
        log_format=log_format
    )

    setup_console_handler(
        cfg=cfg,
        name=script_logger_name,
        log_level=log_level,
        log_format=log_format
    )

    if debug_pathfilename is not None:
        setup_file_handler(
            cfg=cfg,
            name='skyllh',
            filename=debug_pathfilename,
            log_format=log_format,
            log_level=logging.DEBUG
        )
        setup_file_handler(
            cfg=cfg,
            name=script_logger_name,
            filename=debug_pathfilename,
            log_format=log_format,
            log_level=logging.DEBUG
        )

    script_logger = get_logger(script_logger_name)

    return script_logger



# -*- coding: utf-8 -*-

from skyllh.datasets.i3 import (
    PublicData_10y_ps,
    PublicData_10y_ps_wMC,
)

data_samples = {
    'PublicData_10y_ps':
        PublicData_10y_ps,
    'PublicData_10y_ps_wMC':
        PublicData_10y_ps_wMC,
}



# -*- coding: utf-8 -*-
# Author: Dr. Martin Wolf <mail@martin-wolf.org>

import numpy as np

from skyllh.core.dataset import (
    DatasetCollection,
    DatasetOrigin,
    WGETDatasetTransfer,
)
from skyllh.i3.dataset import (
    I3Dataset,
)


def create_dataset_collection(
        cfg,
        base_path=None,
        sub_path_fmt=None,
):
    """Defines the dataset collection for IceCube's 10-year
    point-source public data, which is available at
    http://icecube.wisc.edu/data-releases/20210126_PS-IC40-IC86_VII.zip

    Parameters
    ----------
    cfg : instance of Config
        The instance of Config holding the local configuration.
    base_path : str | None
        The base path of the data files. The actual path of a data file is
        assumed to be of the structure <base_path>/<sub_path>/<file_name>.
        If None, use the default path ``cfg['repository']['base_path']``.
    sub_path_fmt : str | None
        The sub path format of the data files of the public data sample.
        If None, use the default sub path format
        'icecube_10year_ps'.

    Returns
    -------
    dsc : DatasetCollection
        The dataset collection containing all the seasons as individual
        I3Dataset objects.
    """
    # Define the version of the data sample (collection).
    (version, verqualifiers) = (1, dict(p=0))

    # Define the default sub path format.
    default_sub_path_fmt = 'icecube_10year_ps'

    # We create a dataset collection that will hold the individual seasonal
    # public data datasets (all of the same version!).
    dsc = DatasetCollection('Public Data 10-year point-source')

    dsc.description = r"""
    The events contained in this release correspond to the IceCube's
    time-integrated point source search with 10 years of data [2]. Please refer
    to the description of the sample and known changes in the text at [1].

    The data contained in this release of IceCube’s point source sample shows
    evidence of a cumulative excess of events from four sources (NGC 1068,
    TXS 0506+056, PKS 1424+240, and GB6 J1542+6129) from a catalogue of 110
    potential sources. NGC 1068 gives the largest excess and is coincidentally
    the hottest spot in the full Northern sky search [1].

    Data from IC86-2012 through IC86-2014 used in [2] use an updated selection
    and reconstruction compared to the 7 year time-integrated search [3] and the
    detection of the 2014-2015 neutrino flare from the direction of
    TXS 0506+056 [4]. The 7 year and 10 year versions of the sample show
    overlaps of between 80 and 90%.

    An a posteriori cross check of the updated sample has been performed on
    TXS 0506+056 showing two previously-significant cascade-like events removed
    in the newer sample. These two events occur near the blazar's position
    during the TXS flare and give large reconstructed energies, but are likely
    not well-modeled by the track-like reconstructions included in this
    selection. While the events are unlikely to be track-like, their
    contribution to previous results has been handled properly.

    While the significance of the 2014-2015 TXS 0505+56 flare has decreased from
    p=7.0e-5 to 8.1e-3, the change is a result of changes to the sample and not
    of increased data. No problems have been identified with the previously
    published results and since we have no reason a priori to prefer the new
    sample over the old sample, these results do not supercede those in [4].

    This release contains data beginning in 2008 (IC40) until the spring of 2018
    (IC86-2017). This release duplicates and supplants previously released data
    from 2012 and earlier. Events from this release cannot be combined with any
    other releases

    --------------------------
    # Experimental data events
    --------------------------
    The "events" folder contains the events observed in the 10 year sample of
    IceCube's point source neutrino selection. Each file corresponds to a single
    season of IceCube datataking, including roughly one year of data. For each
    event, reconstructed particle information is included.

    - MJD: The MJD time (ut1) of the event interaction given to 1e-8 days,
    corresponding to roughly millisecond precision.

    - log10(E/GeV): The reconstructed energy of a muon passing through the
    detector. The reconstruction follows the prescription for unfolding the
    given in Section 8 of [5].

    - AngErr[deg]: The estimated angular uncertainty on the reconstructed
    direction given in degrees. The angular uncertainty is assumed to be
    symmetric in azimuth and zenith and is used to calculate the signal spatial
    probabilities for each event following the procedure given in [6]. The
    errors are calibrated using simulated events so that they provide correct
    coverage for an E^{-2} power law flux. This sample assumes a lower limit on
    the estimated angular uncertainty of 0.2 degrees.

    - RA[deg], Dec[deg]: The right ascension and declination (J2000)
    corresponding to the particle's reconstructed origin. Given in degrees.

    - Azimuth[deg], Zenith[deg]: The local coordinates of the particle's
    reconstructed origin.

    The local coordinates may be necessary when searching for transient
    phenomena on timescales shorter than 1 day due to non-uniformity in the
    detector's response as a function of azimuth. In these cases, we recommend
    scrambling events in time, then using the local coordinates and time to
    calculate new RA and Dec values.

    Note that during the preparation of this data release, one duplicated event
    was discovered in the IC86-2015 season. This event has not contributed to
    any significant excesses.

    -----------------
    # Detector uptime
    -----------------
    In order to properly account for detector uptime, IceCube maintains
    "good run lists". These contain information about "good runs", periods of
    datataking useful for analysis. Data may be marked unusable for various
    reasons, including major construction or upgrade work, calibration runs, or
    other anomalies. The "uptime" folder contains lists of the good runs for
    each season.

    - MJD_start[days], MJD_stop[days]: The start and end times for each good run

    -------------------------------
    # Instrument response functions
    -------------------------------
    In order to best model the response of the IceCube detector to a given
    signal, Monte Carlo simulations are produced for each detector
    configuration. Events are sampled from these simulations to model the
    response of point sources from an arbitrary source and spectrum.

    We provide several binned responses for the detector in the "irfs" folder
    of this data release.

    -----------------
    # Effective Areas
    -----------------
    The effective area is a property of the detector and selection which, when
    convolved with a flux model, gives the expected rate of events in the
    detector. Here we release the muon neutrino effective areas for each season
    of data.

    The effective areas are averaged over bins using simulated muon neutrino
    events ranging from 100 GeV to 100 PeV. Because the response varies widely
    in both energy and declination, we provide the tabulated response in these
    two dimensions. Due to IceCube's unique position at the south pole, the
    effective area is uniform in right ascension for timescales longer than
    1 day. It varies by about 10% as a function of azimuth, an effect which may
    be important for shorter timescales. While the azimuthal effective areas are
    not included here, they are included in IceCube's internal analyses.
    These may be made available upon request.

    Tabulated versions of the effective area are included in csv files in the
    "irfs" folder. Plotted versions are included as pdf files in the same
    location. Because the detector configuration and selection were unchanged
    after the IC86-2012 season, the effective area for this season should be
    used for IC86-2012 through IC86-2017.

    - log10(E_nu/GeV)_min, log10(E_nu/GeV)_max: The minimum and maximum of the
    energy bin used to caclulate the average effective area. Note that this uses
    the neutrino's true energy and not the reconstructed muon energy.

    - Dec_nu_min[deg], Dec_nu_max[deg]: The minimum and maximum of the
    declination of the neutrino origin. Again, note that this is the true
    direction of the neutrino and not the reconstructed muon direction.

    - A_Eff[cm^2]: The average effective area across a bin.

    -------------------
    # Smearing Matrices
    -------------------
    IceCube has a nontrivial smearing matrix with correlations between the
    directional uncertainty, the point spread function, and the reconstructed
    muon energy. To provide the most complete set of information, we include
    tables of these responses for each season from IC40 through IC86-2012.
    Seasons after IC86-2012 reuse that season's response functions.

    The included smearing matrices take the form of 5D tables mapping a
    (E_nu, Dec_nu) bin in effective area to a 3D matrix of (E, PSF, AngErr).
    The contents of each 3D matrix bin give the fractional count of simulated
    events within the bin relative to all events in the (E_nu, Dec_nu) bin.

    Fractional_Counts = [Events in (E_nu, Dec_nu, E, PSF, AngErr)] /
                        [Events in (E_nu, Dec_nu)]

    The simulations statistics, while large enough for direct sampling, are
    limited when producing these tables, ranging from just 621,858 simulated
    events for IC40 to 11,595,414 simulated events for IC86-2012. In order to
    reduce statistical uncertainties in each 5D bin, bins are selected in each
    (E_nu, Dec_nu) bin independently. The bin edges are given in the smearing
    matrix files. All locations not given have a Fractional_Counts of 0.

    - log10(E_nu/GeV)_min, log10(E_nu/GeV)_max: The minimum and maximum of the
    energy bin used to caclulate the average effective area. Note that this uses
    the neutrino's true energy and not the reconstructed muon energy.

    - Dec_nu_min[deg], Dec_nu_max[deg]: The minimum and maximum of the
    declination of the neutrino origin. Again, note that this is the true
    direction of the neutrino and not the reconstructed muon direction.

    - log10(E/GeV): The reconstructed energy of a muon passing through the
    detector. The reconstruction follows the prescription for unfolding the
    given in Section 8 of [5].

    - PSF_min[deg], PSF_max[deg]: The minimum and maximum of the true angle
    between the neutrino origin and the reconstructed muon direction.

    - AngErr_min[deg], AngErr_max[deg]: The estimated angular uncertainty on the
    reconstructed direction given in degrees. The angular uncertainty is assumed
    to be symmetric in azimuth and zenith and is used to calculate the signal
    spatial probabilities for each event following the procedure given in [6].
    The errors are calibrated so that they provide correct coverage for an
    E^{-2} power law flux. This sample assumes a lower limit on the estimated
    angular uncertainty of 0.2 degrees.

    - Fractional_Counts: The fraction of simulated events falling within each
    5D bin relative to all events in the (E_nu, Dec_nu) bin.

    ------------
    # References
    ------------
    [1] IceCube Data for Neutrino Point-Source Searches: Years 2008-2018,
        [ArXiv link](https://arxiv.org/abs/2101.09836)
    [2] Time-integrated Neutrino Source Searches with 10 years of IceCube Data,
        Phys. Rev. Lett. 124, 051103 (2020)
    [3] All-sky search for time-integrated neutrino emission from astrophysical
        sources with 7 years of IceCube data,
        Astrophys. J., 835 (2017) no. 2, 151
    [4] Neutrino emission from the direction of the blazar TXS 0506+056 prior to
        the IceCube-170922A alert,
        Science 361, 147-151 (2018)
    [5] Energy Reconstruction Methods in the IceCube Neutrino Telescope,
        JINST 9 (2014), P03009
    [6] Methods for point source analysis in high energy neutrino telescopes,
        Astropart.Phys.29:299-305,2008

    -------------
    # Last Update
    -------------
    28 January 2021
    """

    # Define the origin of the dataset.
    origin = DatasetOrigin(
        base_path='data-releases',
        sub_path='',
        filename='20210126_PS-IC40-IC86_VII.zip',
        host='icecube.wisc.edu',
        transfer_func=WGETDatasetTransfer(protocol='http').transfer,
        post_transfer_func=WGETDatasetTransfer.post_transfer_unzip,
    )

    # Define the common keyword arguments for all data sets.
    ds_kwargs = dict(
        cfg=cfg,
        livetime=None,
        version=version,
        verqualifiers=verqualifiers,
        base_path=base_path,
        default_sub_path_fmt=default_sub_path_fmt,
        sub_path_fmt=sub_path_fmt,
        origin=origin,
    )

    grl_field_name_renaming_dict = {
        'MJD_start[days]': 'start',
        'MJD_stop[days]': 'stop',
    }

    # Define the datasets for the different seasons.
    # For the declination and energy binning we use the same binning as was
    # used in the original point-source analysis using the PointSourceTracks
    # dataset.

    # ---------- IC40 ----------------------------------------------------------
    IC40 = I3Dataset(
        name='IC40',
        exp_pathfilenames='events/IC40_exp.csv',
        mc_pathfilenames=None,
        grl_pathfilenames='uptime/IC40_exp.csv',
        **ds_kwargs,
    )
    IC40.grl_field_name_renaming_dict = grl_field_name_renaming_dict
    IC40.add_aux_data_definition(
        'eff_area_datafile', 'irfs/IC40_effectiveArea.csv')
    IC40.add_aux_data_definition(
        'smearing_datafile', 'irfs/IC40_smearing.csv')

    sin_dec_bins = np.unique(np.concatenate([
        np.linspace(-1., -0.25, 10 + 1),
        np.linspace(-0.25, 0.0, 10 + 1),
        np.linspace(0.0, 1., 10 + 1),
    ]))
    IC40.define_binning('sin_dec', sin_dec_bins)

    energy_bins = np.arange(2., 9.5 + 0.01, 0.125)
    IC40.define_binning('log_energy', energy_bins)

    # ---------- IC59 ----------------------------------------------------------
    IC59 = I3Dataset(
        name='IC59',
        exp_pathfilenames='events/IC59_exp.csv',
        mc_pathfilenames=None,
        grl_pathfilenames='uptime/IC59_exp.csv',
        **ds_kwargs,
    )
    IC59.grl_field_name_renaming_dict = grl_field_name_renaming_dict
    IC59.add_aux_data_definition(
        'eff_area_datafile', 'irfs/IC59_effectiveArea.csv')
    IC59.add_aux_data_definition(
        'smearing_datafile', 'irfs/IC59_smearing.csv')

    sin_dec_bins = np.unique(np.concatenate([
        np.linspace(-1., -0.95, 2 + 1),
        np.linspace(-0.95, -0.25, 25 + 1),
        np.linspace(-0.25, 0.05, 15 + 1),
        np.linspace(0.05, 1., 10 + 1),
    ]))
    IC59.define_binning('sin_dec', sin_dec_bins)

    energy_bins = np.arange(2., 9.5 + 0.01, 0.125)
    IC59.define_binning('log_energy', energy_bins)

    # ---------- IC79 ----------------------------------------------------------
    IC79 = I3Dataset(
        name='IC79',
        exp_pathfilenames='events/IC79_exp.csv',
        mc_pathfilenames=None,
        grl_pathfilenames='uptime/IC79_exp.csv',
        **ds_kwargs,
    )
    IC79.grl_field_name_renaming_dict = grl_field_name_renaming_dict
    IC79.add_aux_data_definition(
        'eff_area_datafile', 'irfs/IC79_effectiveArea.csv')
    IC79.add_aux_data_definition(
        'smearing_datafile', 'irfs/IC79_smearing.csv')

    sin_dec_bins = np.unique(np.concatenate([
        np.linspace(-1., -0.75, 10 + 1),
        np.linspace(-0.75, 0., 15 + 1),
        np.linspace(0., 1., 20 + 1),
    ]))
    IC79.define_binning('sin_dec', sin_dec_bins)

    energy_bins = np.arange(2., 9.5 + 0.01, 0.125)
    IC79.define_binning('log_energy', energy_bins)

    # ---------- IC86-I --------------------------------------------------------
    IC86_I = I3Dataset(
        name='IC86_I',
        exp_pathfilenames='events/IC86_I_exp.csv',
        mc_pathfilenames=None,
        grl_pathfilenames='uptime/IC86_I_exp.csv',
        **ds_kwargs,
    )
    IC86_I.grl_field_name_renaming_dict = grl_field_name_renaming_dict
    IC86_I.add_aux_data_definition(
        'eff_area_datafile', 'irfs/IC86_I_effectiveArea.csv')
    IC86_I.add_aux_data_definition(
        'smearing_datafile', 'irfs/IC86_I_smearing.csv')

    b = np.sin(np.radians(-5.))  # North/South transition boundary.
    sin_dec_bins = np.unique(np.concatenate([
        np.linspace(-1., -0.2, 10 + 1),
        np.linspace(-0.2, b, 4 + 1),
        np.linspace(b, 0.2, 5 + 1),
        np.linspace(0.2, 1., 10),
    ]))
    IC86_I.define_binning('sin_dec', sin_dec_bins)

    energy_bins = np.arange(1., 10.5 + 0.01, 0.125)
    IC86_I.define_binning('log_energy', energy_bins)

    # ---------- IC86-II -------------------------------------------------------
    IC86_II = I3Dataset(
        name='IC86_II',
        exp_pathfilenames='events/IC86_II_exp.csv',
        mc_pathfilenames=None,
        grl_pathfilenames='uptime/IC86_II_exp.csv',
        **ds_kwargs,
    )
    IC86_II.grl_field_name_renaming_dict = grl_field_name_renaming_dict
    IC86_II.add_aux_data_definition(
        'eff_area_datafile', 'irfs/IC86_II_effectiveArea.csv')
    IC86_II.add_aux_data_definition(
        'smearing_datafile', 'irfs/IC86_II_smearing.csv')

    sin_dec_bins = np.unique(np.concatenate([
        np.linspace(-1., -0.93, 4 + 1),
        np.linspace(-0.93, -0.3, 10 + 1),
        np.linspace(-0.3, 0.05, 9 + 1),
        np.linspace(0.05, 1., 18 + 1),
    ]))
    IC86_II.define_binning('sin_dec', sin_dec_bins)

    energy_bins = np.arange(0.5, 9.5 + 0.01, 0.125)
    IC86_II.define_binning('log_energy', energy_bins)

    # ---------- IC86-III ------------------------------------------------------
    IC86_III = I3Dataset(
        name='IC86_III',
        exp_pathfilenames='events/IC86_III_exp.csv',
        mc_pathfilenames=None,
        grl_pathfilenames='uptime/IC86_III_exp.csv',
        **ds_kwargs,
    )
    IC86_III.grl_field_name_renaming_dict = grl_field_name_renaming_dict
    IC86_III.add_aux_data_definition(
        'eff_area_datafile', 'irfs/IC86_II_effectiveArea.csv')
    IC86_III.add_aux_data_definition(
        'smearing_datafile', 'irfs/IC86_II_smearing.csv')

    IC86_III.add_binning_definition(
        IC86_II.get_binning_definition('sin_dec'))
    IC86_III.add_binning_definition(
        IC86_II.get_binning_definition('log_energy'))

    # ---------- IC86-IV -------------------------------------------------------
    IC86_IV = I3Dataset(
        name='IC86_IV',
        exp_pathfilenames='events/IC86_IV_exp.csv',
        mc_pathfilenames=None,
        grl_pathfilenames='uptime/IC86_IV_exp.csv',
        **ds_kwargs,
    )
    IC86_IV.grl_field_name_renaming_dict = grl_field_name_renaming_dict
    IC86_IV.add_aux_data_definition(
        'eff_area_datafile', 'irfs/IC86_II_effectiveArea.csv')
    IC86_IV.add_aux_data_definition(
        'smearing_datafile', 'irfs/IC86_II_smearing.csv')

    IC86_IV.add_binning_definition(
        IC86_II.get_binning_definition('sin_dec'))
    IC86_IV.add_binning_definition(
        IC86_II.get_binning_definition('log_energy'))

    # ---------- IC86-V --------------------------------------------------------
    IC86_V = I3Dataset(
        name='IC86_V',
        exp_pathfilenames='events/IC86_V_exp.csv',
        mc_pathfilenames=None,
        grl_pathfilenames='uptime/IC86_V_exp.csv',
        **ds_kwargs,
    )
    IC86_V.grl_field_name_renaming_dict = grl_field_name_renaming_dict
    IC86_V.add_aux_data_definition(
        'eff_area_datafile', 'irfs/IC86_II_effectiveArea.csv')
    IC86_V.add_aux_data_definition(
        'smearing_datafile', 'irfs/IC86_II_smearing.csv')

    IC86_V.add_binning_definition(
        IC86_II.get_binning_definition('sin_dec'))
    IC86_V.add_binning_definition(
        IC86_II.get_binning_definition('log_energy'))

    # ---------- IC86-VI -------------------------------------------------------
    IC86_VI = I3Dataset(
        name='IC86_VI',
        exp_pathfilenames='events/IC86_VI_exp.csv',
        mc_pathfilenames=None,
        grl_pathfilenames='uptime/IC86_VI_exp.csv',
        **ds_kwargs,
    )
    IC86_VI.grl_field_name_renaming_dict = grl_field_name_renaming_dict
    IC86_VI.add_aux_data_definition(
        'eff_area_datafile', 'irfs/IC86_II_effectiveArea.csv')
    IC86_VI.add_aux_data_definition(
        'smearing_datafile', 'irfs/IC86_II_smearing.csv')

    IC86_VI.add_binning_definition(
        IC86_II.get_binning_definition('sin_dec'))
    IC86_VI.add_binning_definition(
        IC86_II.get_binning_definition('log_energy'))

    # ---------- IC86-VII ------------------------------------------------------
    IC86_VII = I3Dataset(
        name='IC86_VII',
        exp_pathfilenames='events/IC86_VII_exp.csv',
        mc_pathfilenames=None,
        grl_pathfilenames='uptime/IC86_VII_exp.csv',
        **ds_kwargs,
    )
    IC86_VII.grl_field_name_renaming_dict = grl_field_name_renaming_dict
    IC86_VII.add_aux_data_definition(
        'eff_area_datafile', 'irfs/IC86_II_effectiveArea.csv')
    IC86_VII.add_aux_data_definition(
        'smearing_datafile', 'irfs/IC86_II_smearing.csv')

    IC86_VII.add_binning_definition(
        IC86_II.get_binning_definition('sin_dec'))
    IC86_VII.add_binning_definition(
        IC86_II.get_binning_definition('log_energy'))

    # ---------- IC86-II-VII ---------------------------------------------------
    ds_list = [
        IC86_II,
        IC86_III,
        IC86_IV,
        IC86_V,
        IC86_VI,
        IC86_VII,
    ]
    IC86_II_VII = I3Dataset(
        name='IC86_II-VII',
        exp_pathfilenames=I3Dataset.get_combined_exp_pathfilenames(ds_list),
        mc_pathfilenames=None,
        grl_pathfilenames=I3Dataset.get_combined_grl_pathfilenames(ds_list),
        **ds_kwargs,
    )
    IC86_II_VII.grl_field_name_renaming_dict = grl_field_name_renaming_dict
    IC86_II_VII.add_aux_data_definition(
        'eff_area_datafile',
        IC86_II.get_aux_data_definition('eff_area_datafile'))

    IC86_II_VII.add_aux_data_definition(
        'smearing_datafile',
        IC86_II.get_aux_data_definition('smearing_datafile'))

    IC86_II_VII.add_binning_definition(
        IC86_II.get_binning_definition('sin_dec'))
    IC86_II_VII.add_binning_definition(
        IC86_II.get_binning_definition('log_energy'))

    # --------------------------------------------------------------------------

    dsc.add_datasets((
        IC40,
        IC59,
        IC79,
        IC86_I,
        IC86_II,
        IC86_III,
        IC86_IV,
        IC86_V,
        IC86_VI,
        IC86_VII,
        IC86_II_VII,
    ))

    dsc.set_exp_field_name_renaming_dict({
        'MJD[days]':    'time',
        'log10(E/GeV)': 'log_energy',
        'AngErr[deg]':  'ang_err',
        'RA[deg]':      'ra',
        'Dec[deg]':     'dec',
        'Azimuth[deg]': 'azi',
        'Zenith[deg]':  'zen',
    })

    def convert_deg2rad(data):
        exp = data.exp
        exp['ang_err'] = np.deg2rad(exp['ang_err'])
        exp['ra'] = np.deg2rad(exp['ra'])
        exp['dec'] = np.deg2rad(exp['dec'])
        exp['azi'] = np.deg2rad(exp['azi'])
        exp['zen'] = np.deg2rad(exp['zen'])

    dsc.add_data_preparation(convert_deg2rad)

    return dsc



# -*- coding: utf-8 -*-

import abc
import copy
import functools
import inspect
import numpy as np
import sys

from collections import OrderedDict

from skyllh.core.display import INDENTATION_WIDTH


class PyQualifier(
        object,
        metaclass=abc.ABCMeta):
    """This is the abstract base class for any Python qualifier class.
    An object can get qualified by calling a PyQualifier instance with that
    object. The PyQualifier class will be added to the ``__pyqualifiers__``
    attribute of the object.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __call__(self, obj):
        """Declares the given Python object to be qualified with the
        PyQualifier class of `self`. It adds the class of `self` to the
        `__pyqualifiers__` tuple of the given object.

        Parameters
        ----------
        obj : object
            The Python object that should get qualified.

        Returns
        -------
        obj : object
            The given object, but modified to be declared for this Python
            qualifier.
        """
        if not hasattr(obj, '__pyqualifiers__'):
            setattr(obj, '__pyqualifiers__', ())

        obj.__pyqualifiers__ += (self.__class__,)

        return obj

    def check(self, obj):
        """Checks if the given Python object is declared with the PyQualifier
        class of the `self` object.

        Parameters
        ----------
        obj : object
            The Python object to check.

        Returns
        -------
        check : bool
            The check result. `True` if the object is declared for this Python
            qualifier, and `False` otherwise.
        """
        if not hasattr(obj, '__pyqualifiers__'):
            return False

        if self.__class__ in obj.__pyqualifiers__:
            return True

        return False


class ConstPyQualifier(
        PyQualifier):
    """This class defines a PyQualifier for constant Python objects.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


const = ConstPyQualifier()


def get_class_of_func(f):
    """Determines the class object that defined the given method or function
    ``f``.

    Parameters
    ----------
    f : function | method
        The function or method whose parent class should be determined.

    Returns
    -------
    cls : class | None
        The class object which defines the given function or method. ``None``
        is returned when no class could be determined.
    """
    if isinstance(f, functools.partial):
        return get_class_of_func(f.func)

    if inspect.ismethod(f) or\
        ((inspect.isbuiltin(f)) and
         (getattr(f, '__self__', None) is not None) and
         (getattr(f.__self__, '__class__', None))):
        for cls in inspect.getmro(f.__self__.__class__):
            if hasattr(cls, '__dict__') and (f.__name__ in cls.__dict__):
                return cls
        # Fall back to normal function evaluation.
        f = getattr(f, '__func__', f)

    if inspect.isfunction(f):
        cls = getattr(
            inspect.getmodule(f),
            f.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)[0],
            None)
        if isinstance(cls, type):
            return cls

    # Handle special descriptor objects.
    cls = getattr(f, '__objclass__', None)
    return cls


def typename(t):
    """Returns the name of the given type ``t``.
    """
    return t.__name__


def classname(obj):
    """Returns the name of the class of the class instance ``obj``.
    """
    return typename(type(obj))


def module_classname(obj):
    """Returns the module and class name of the given instance ``obj``.
    """
    return f'{obj.__module__}.{classname(obj)}'


def module_class_method_name(obj, meth_name):
    """Returns the module, class, and method name of the given instance ``obj``.

    Parameters
    ----------
    obj : instance of object
        The object instance.
    meth_name : str
        The name of the method.
    """
    return f'{module_classname(obj)}.{meth_name}'


def get_byte_size_prefix(size):
    """Determines the biggest size prefix for the given size in bytes such that
    the new size is still greater one.

    Parameters
    ----------
    size : int
        The size in bytes.

    Returns
    -------
    newsize : float
        The new byte size accounting for the byte prefix.
    prefix : str
        The biggest byte size prefix.
    """
    prefix_factor_list = [
        ('', 1), ('K', 1024), ('M', 1024**2), ('G', 1024**3), ('T', 1024**4)]

    prefix_idx = 0
    for (prefix, factor) in prefix_factor_list[1:]:
        if size / factor < 1:
            break
        prefix_idx += 1

    (prefix, factor) = prefix_factor_list[prefix_idx]
    newsize = size / factor

    return (newsize, prefix)


def getsizeof(objects):
    """Determines the size in bytes the given objects have in memory.
    If an object is a sequence, the size of the elements of the sequence will
    be estimated as well and added to the result. This does not account for the
    multiple occurence of the same object.

    Parameters
    ----------
    objects : sequence of instances of object | instance of object.

    Returns
    -------
    memsize : int
        The memory size in bytes of the given objects.
    """
    if not issequence(objects):
        objects = [objects]

    memsize = 0
    for obj in objects:
        if issequence(obj):
            memsize += getsizeof(obj)
        else:
            memsize += sys.getsizeof(obj)

    return memsize


def issequence(obj):
    """Checks if the given object ``obj`` is a sequence or not. The definition of
    a sequence in this case is, that the function ``len`` is defined for the
    object.

    .. note::

        A str object is NOT considered a sequence!

    Returns
    -------
    check : bool
        ``True`` if the given object is a sequence. ``False`` if the given
        object is an instance of str or not a sequence.

    """
    if isinstance(obj, str):
        return False

    try:
        len(obj)
    except TypeError:
        return False

    return True


def issequenceof(obj, T, pyqualifiers=None):
    """Checks if the given object ``obj`` is a sequence with items being
    instances of type ``T``, possibly qualified with the given Python
    qualifiers.

    Parameters
    ----------
    obj : instance of object
        The Python object to check.
    T : type | tuple of types
        The type each item of the sequence should be. If a tuple of types is
        given, each item can be one of the given types.
    pyqualifiers : instance of PyQualifier |
            sequence of instances of PyQualifier | None
        One or more instances of PyQualifier. Each instance acts as a filter
        for each item.
        If any of the filters return `False` for any of the items, this check
        returns `False`. If set to `Ǹone`, no filters are applied.

    Returns
    -------
    check : bool
        The result of the check.
    """
    if pyqualifiers is None:
        pyqualifiers = tuple()
    elif not issequence(pyqualifiers):
        pyqualifiers = (pyqualifiers,)

    if not issequence(obj):
        return False

    for item in obj:
        if not isinstance(item, T):
            return False
        for pyqualifier in pyqualifiers:
            if not pyqualifier.check(item):
                return False

    return True


def issequenceofsubclass(obj, T):
    """Checks if the given object ``obj`` is a sequence with items being
    sub-classes of class T.

    Parameters
    ----------
    obj : instance of object
        The object to check.
    T : class
        The base class of the items of the given object.

    Returns
    -------
    check : bool
        ``True`` if the given object is a sequence of instances which are
        sub-classes of class ``T``. ``False`` if ``obj`` is not a sequence or
        any item is not a sub-class of class ``T``.
    """
    if not issequence(obj):
        return False

    for item in obj:
        if not issubclass(item, T):
            return False

    return True


def isproperty(obj, name):
    """Checks if the given attribute is of type property. The attribute must
    exist in ``obj``.

    Parameters
    ----------
    obj : object
        The Python object whose attribute to check for being a property.
    name : str
        The name of the attribute.

    Returns
    -------
    check : bool
        True if the given attribute is of type property, False otherwise.

    Raises
    ------
    AttributeError
        If the given attribute is not an attribute of the class of ``obj``.
    """
    attr = type(obj).__class__.__getattribute__(type(obj), name)
    return isinstance(attr, property)


def func_has_n_args(func, n):
    """Checks if the given function `func` has `n` arguments.

    Parameters
    ----------
    func : callable
        The function to check.
    n : int
        The number of arguments the function must have.

    Returns
    -------
    check : bool
        True if the given function has `n` arguments. False otherwise.
    """
    check = (len(inspect.signature(func).parameters) == n)
    return check


def bool_cast(v, errmsg):
    """Casts the given value to a boolean value. If the cast is impossible, a
    TypeError is raised with the given error message.
    """
    try:
        v = bool(v)
    except Exception:
        raise TypeError(errmsg)

    return v


def int_cast(v, errmsg, allow_None=False):
    """Casts the given value to an integer value. If the cast is impossible, a
    TypeError is raised with the given error message. If `allow_None` is set to
    `True` the value `v` can also be `None`.
    """
    if allow_None and v is None:
        return v

    try:
        v = int(v)
    except Exception:
        raise TypeError(errmsg)

    return v


def float_cast(v, errmsg, allow_None=False):
    """Casts the given value to a float. If the cast is impossible, a TypeError
    is raised with the given error message. If `allow_None` is set to `True`
    the value `v` can also be `None`.

    Parameters
    ----------
    v : to_float_castable object | sequence of to_float_castable objects
        The object that should get casted to a float. This can also be a
        sequence of objects that should get casted to floats.
    errmsg : str
        The error message in case the cast failed.
    allow_None : bool
        Flag if ``None`` is allowed as value for v. If yes, the casted result is
        ``None``.

    Returns
    -------
    float | None | list of float and or None
        The float / ``None`` value casted from ``v``. If a sequence of objects
        was provided, a list of casted values is returned.
    """
    # Define cast function for a single object.
    def _obj_float_cast(v, errmsg, allow_None):
        if allow_None and v is None:
            return v

        try:
            v = float(v)
        except Exception:
            raise TypeError(errmsg)

        return v

    if issequence(v):
        float_list = []
        for el_v in v:
            float_list.append(_obj_float_cast(el_v, errmsg, allow_None))
        return float_list

    return _obj_float_cast(v, errmsg, allow_None)


def str_cast(v, errmsg, allow_None=False):
    """Casts the given value to a str object.
    If the cast is impossible, a TypeError is raised with the given error
    message.
    """
    if allow_None and v is None:
        return v

    try:
        v = str(v)
    except Exception:
        raise TypeError(errmsg)

    return v


def list_of_cast(t, v, errmsg):
    """Casts the given value `v` to a list of items of type `t`.
    If the cast is impossible, a TypeError is raised with the given error
    message.
    """
    if isinstance(v, t):
        v = [v]

    if not issequenceof(v, t):
        raise TypeError(errmsg)

    v = list(v)

    return v


def get_smallest_numpy_int_type(values):
    """Returns the smallest numpy integer type that can represent the given
    integer values.

    Parameters
    ----------
    values : int | sequence of int
        The integer value(s) that need to be representable by the returned
        integer type.

    Returns
    -------
    inttype : numpy integer type
        The smallest numpy integer type that can represent the given values.
    """
    values = np.atleast_1d(values)

    vmin = np.min(values)
    vmax = np.max(values)

    if vmin < 0:
        types = [np.int8, np.int16, np.int32, np.int64]
    else:
        types = [np.uint8, np.uint16, np.uint32, np.uint64]

    for inttype in types:
        ii = np.iinfo(inttype)
        if vmin >= ii.min and vmax <= ii.max:
            return inttype

    raise ValueError(
        f'No integer type spans [{vmin}, {vmax}]!')


def get_number_of_float_decimals(value):
    """Determines the number of significant decimals the given float number has.
    The maximum number of supported decimals is 16.

    Parameters
    ----------
    value : float
        The float value whose number of significant decimals should get
        determined.

    Returns
    -------
    decimals : int
        The number of decimals of value which are non-zero.

    Raises
    ------
    ValueError
        If a nan value was provided.
    """
    if np.isnan(value):
        raise ValueError(
            'The provided value is nan!')

    val_str = '{:.16f}'.format(value)
    (val_num_str, val_decs_str) = val_str.split('.', 1)
    for idx in range(len(val_decs_str)-1, -1, -1):
        if int(val_decs_str[idx]) != 0:
            return idx+1

    return 0


def make_dict_hash(d):
    """Creates a hash value for the given dictionary.

    Parameters
    ----------
    d : dict | None
        The dictionary holding (name: value) pairs.
        If set to None, an empty dictionary is used.

    Returns
    -------
    hash : int
        The hash of the dictionary.
    """
    if d is None:
        d = {}

    if not isinstance(d, dict):
        raise TypeError(
            'The d argument must be of type dict!')

    # A note on the ordering of Python dictionary items: The items are ordered
    # internally according to the hash value of their keys. Hence, if we don't
    # insert more dictionary items, the order of the items won't change. Thus,
    # we can just take the items list and make a tuple to create a hash of it.
    # The hash will be the same for two dictionaries having the same items.
    return hash(tuple(d.items()))


class ObjectCollection(
        object):
    """This class provides a collection of objects of a specific type. Objects
    can be added to the collection via the ``add`` method or can be removed
    from the collection via the ``pop`` method. The objects of another object
    collection can be added to this object collection via the ``add`` method as
    well.
    """
    def __init__(self, objs=None, obj_type=None):
        """Constructor of the ObjectCollection class. Must be called by the
        derived class.

        Parameters
        ----------
        objs : instance of obj_type | sequence of obj_type instances | None
            The sequence of objects of type ``obj_type`` with which this
            collection should get initialized with.
        obj_type : type | None
            The type of the objects, which can be added to the collection.
            If set to None, the type will be determined from the given objects.
            If no objects are given, the object type will be `object`.
        """
        if obj_type is None:
            obj_type = object
            if objs is not None:
                if issequence(objs) and len(objs) > 0:
                    obj_type = type(objs[0])
                else:
                    obj_type = type(objs)

        if not issubclass(obj_type, object):
            raise TypeError(
                'The obj_type argument must be a subclass of object!')

        self._obj_type = obj_type
        self._objects = []

        # Add given list of objects.
        if objs is not None:
            if not issequence(objs):
                objs = [objs]
            for obj in objs:
                self.add(obj)

    @property
    def obj_type(self):
        """(read-only) The object type.
        """
        return self._obj_type

    @property
    def objects(self):
        """(read-only) The list of objects of this object collection.
        All objects are of the same type as specified through the ``obj_type``
        property.
        """
        return self._objects

    def __len__(self):
        """Returns the number of objects being in this object collection.
        """
        return len(self._objects)

    def __getitem__(self, key):
        return self._objects[key]

    def __iter__(self):
        return iter(self._objects)

    def __add__(self, other):
        """Implementation to support the operation ``oc = self + other``, where
        ``self`` is this ObjectCollection object and ``other`` something useful
        else. This creates a copy ``oc`` of ``self`` and adds ``other``
        to ``oc``.

        Parameters
        ----------
        other : obj_type | ObjectCollection of obj_type

        Returns
        -------
        oc : ObjectCollection
            The new ObjectCollection object with object from self and other.
        """
        oc = self.copy()
        oc.add(other)
        return oc

    def __str__(self):
        """Pretty string representation of this object collection.
        """
        obj_str = ",\n".join([
            ' '*INDENTATION_WIDTH + str(obj) for obj in self._objects
        ])
        return classname(self) + ": {\n" + obj_str + "\n}"

    def copy(self):
        """Creates a copy of this ObjectCollection. The objects of the
        collection are not copied!
        """
        oc = ObjectCollection(self._obj_type)
        oc._objects = copy.copy(self._objects)
        return oc

    def add(self, obj):
        """Adds the given object, sequence of objects, or object collection to
        this object collection.

        Parameters
        ----------
        obj : obj_type instance | sequence of obj_type |
              ObjectCollection of obj_type
            An instance of ``obj_type`` that should be added to the collection.
            If given an ObjectCollection for objects of type obj_type, it will
            add all objects of the given collection to this collection.

        Returns
        -------
        self : ObjectCollection
            The instance of this ObjectCollection, in order to be able to chain
            several ``add`` calls.
        """
        if issequence(obj) and not isinstance(obj, ObjectCollection):
            obj = ObjectCollection(obj)

        if isinstance(obj, ObjectCollection):
            if not issubclass(obj.obj_type, self.obj_type):
                raise TypeError(
                    'Cannot add objects from ObjectCollection for objects of '
                    f'type "{typename(obj.obj_type)}" to this ObjectCollection '
                    f'for objects of type "{typename(self._obj_type)}"!')
            self._objects.extend(obj.objects)
            return self

        if not isinstance(obj, self._obj_type):
            raise TypeError(
                f'The object of type "{classname(obj)}" cannot be added to the '
                'object collection for objects of type '
                f'"{typename(self._obj_type)}"!')

        self._objects.append(obj)
        return self
    __iadd__ = add

    def index(self, obj):
        """Gets the index of the given object instance within this object
        collection.

        Parameters
        ----------
        obj : obj_type instance
            The instance of obj_type whose index should get retrieved.

        Returns
        -------
        idx : int
            The index of the object within this object collection.
        """
        return self._objects.index(obj)

    def pop(self, index=None):
        """Removes and returns the object at the given index (default last).
        Raises IndexError if the collection is empty or index is out of range.

        Parameters
        ----------
        index : int | None
            The index of the object to remove. If set to None, the index of the
            last object is used.

        Returns
        -------
        obj : obj_type
            The removed object.
        """
        if index is None:
            index = len(self._objects) - 1
        obj = self._objects.pop(index)
        return obj


class NamedObjectCollection(ObjectCollection):
    """This class provides a collection of objects, which have a name. Access
    via the object name is efficient because the index of each object is
    tracked w.r.t. its name.
    """
    def __init__(self, objs=None, obj_type=None, **kwargs):
        """Creates a new NamedObjectCollection instance. Must be called by the
        derived class.

        Parameters
        ----------
        objs : instance of obj_type | sequence of instances of obj_type | None
            The sequence of objects of type ``obj_type`` with which this
            collection should get initialized with.
        obj_type : type
            The type of the objects, which can be added to the collection.
            This type must have an attribute named ``name``.
        """
        self._obj_name_to_idx = OrderedDict()

        # The ObjectCollection class will call the add method to add individual
        # objects. This will update the _obj_name_to_idx attribute.
        super().__init__(
            objs=objs,
            obj_type=obj_type,
            **kwargs)

        if not hasattr(self.obj_type, 'name'):
            raise TypeError(
                f'The object type "{typename(self.obj_type)}" has no '
                'attribute named "name"!')

    @property
    def name_list(self):
        """(read-only) The list of the names of all the objects of this
        NamedObjectCollection instance.
        The order of this list of names is preserved to the order objects were
        added to this collection.
        """
        return list(self._obj_name_to_idx.keys())

    def _create_obj_name_to_idx_dict(self, start=None, end=None):
        """Creates the dictionary {obj.name: index} for object in the interval
        [`start`, `end`).

        Parameters
        ----------
        start : int | None
            The object start index position, which is inclusive.
        end : int | None
            The object end index position, which is exclusive.

        Returns
        -------
        obj_name_to_idx : dict
            The dictionary {obj.name: index}.
        """
        if start is None:
            start = 0

        return OrderedDict([
            (o.name, start+idx)
            for (idx, o) in enumerate(self._objects[start:end])
        ])

    def __contains__(self, name):
        """Returns ``True`` if an object of the given name exists in this
        NamedObjectCollection instance, ``False`` otherwise.

        Parameters
        ----------
        name : str
            The name of the object.

        Returns
        -------
        check : bool
            ``True`` if an object of name ``name`` exists in this
            NamedObjectCollection instance, ``False`` otherwise.
        """
        return name in self._obj_name_to_idx

    def __getitem__(self, key):
        """Returns an object based on its name or index.

        Parameters
        ----------
        key : str | int
            The object identification. Either its name or its index position
            within the object collection.

        Returns
        -------
        obj : instance of obj_type
            The requested object.

        Raises
        ------
        KeyError
            If the given object is not found within this object collection.
        """
        if isinstance(key, str):
            key = self.get_index_by_name(key)
        return super().__getitem__(key)

    def add(self, obj):
        """Adds the given object to this named object collection.

        Parameters
        ----------
        obj : obj_type instance | NamedObjectCollection of obj_type
            An instance of ``obj_type`` that should be added to this named
            object collection.
            If a NamedObjectCollection instance for objects of type ``obj_type``
            is given, all objects of that given collection will be added to this
            named object collection.

        Returns
        -------
        self : NamedObjectCollection
            The instance of this NamedObjectCollection, in order to be able to
            chain several ``add`` calls.
        """
        n_objs = len(self)

        super().add(obj)

        self._obj_name_to_idx.update(
            self._create_obj_name_to_idx_dict(n_objs))

        return self
    __iadd__ = add

    def get_index_by_name(self, name):
        """Gets the index of the object with the given name within this named
        object collection.

        Parameters
        ----------
        name : str
            The name of the object whose index should get retrieved.

        Returns
        -------
        idx : int
            The index of the object within this named object collection.
        """
        return self._obj_name_to_idx[name]

    def pop(self, index=None):
        """Removes and returns the object at the given index (default last).
        Raises IndexError if the collection is empty or index is out of range.

        Parameters
        ----------
        index : int | str | None
            The index of the object to remove. If set to None, the index of the
            last object is used.
            If a str instance is given, it specifies the name of the object.

        Returns
        -------
        obj : obj_type instance
            The removed object.
        """
        if isinstance(index, str):
            # Get the index of the object given its name.
            index = self.get_index_by_name(index)

        obj = super().pop(index)

        # Recreate the object name to index dictionary.
        self._obj_name_to_idx = self._create_obj_name_to_idx_dict()

        return obj



# -*- coding: utf-8 -*-

from astropy import units
import numpy as np

import scipy.interpolate

from skyllh.analyses.i3.publicdata_ps.aeff import (
    load_effective_area_array,
)
from skyllh.core import (
    multiproc,
)
from skyllh.core.binning import (
    BinningDefinition,
)
from skyllh.core.flux_model import (
    FactorizedFluxModel,
)
from skyllh.core.livetime import (
    Livetime,
)
from skyllh.core.py import (
    classname,
    issequence,
)
from skyllh.i3.detsigyield import (
    SingleParamFluxPointLikeSourceI3DetSigYieldBuilder,
    SingleParamFluxPointLikeSourceI3DetSigYield,
)


class PDSingleParamFluxPointLikeSourceI3DetSigYieldBuilder(
        SingleParamFluxPointLikeSourceI3DetSigYieldBuilder,
):
    """This detector signal yield builder class constructs a
    detector signal yield instance for a variable flux model of a single
    parameter, assuming a point-like source.
    It constructs a two-dimensional spline function in sin(dec) and the
    parameter, using a :class:`scipy.interpolate.RectBivariateSpline`.
    Hence, the detector signal yield can vary with the declination and the
    parameter of the flux model.

    It is tailored to the IceCube detector at the South Pole, where the
    effective area depends solely on the zenith angle, and hence on the
    declination, of the source.

    It takes the effective area for the detector signal yield from the auxilary
    detector effective area data file given by the public data.
    """

    def __init__(
            self,
            param_grid,
            spline_order_sinDec=2,
            spline_order_param=2,
            ncpu=None,
            **kwargs,
    ):
        """Creates a new IceCube detector signal yield builder instance for
        a flux model with a single parameter.
        It requires the effective area from the public data, and a parameter
        grid to compute the parameter dependency of the detector signal yield.

        Parameters
        ----------
        param_grid : instance of ParameterGrid
            The instance of ParameterGrid which defines the grid of parameter
            values.
        spline_order_sinDec : int
            The order of the spline function for the logarithmic values of the
            detector signal yield along the sin(dec) axis.
            The default is 2.
        spline_order_param : int
            The order of the spline function for the logarithmic values of the
            detector signal yield along the parameter axis.
            The default is 2.
        ncpu : int | None
            The number of CPUs to utilize. If set to ``None``, global setting
            will take place.
        """
        super().__init__(
            param_grid=param_grid,
            sin_dec_binning=None,
            spline_order_sinDec=spline_order_sinDec,
            spline_order_param=spline_order_param,
            ncpu=ncpu,
            **kwargs)

    def assert_types_of_construct_detsigyield_arguments(
            self,
            shgs,
            **kwargs):
        """Checks the correct types of the arguments for the
        ``construct_detsigyield`` method.
        """
        super().assert_types_of_construct_detsigyield_arguments(
            shgs=shgs,
            **kwargs)

        if not issequence(shgs):
            shgs = [shgs]
        for shg in shgs:
            if not isinstance(shg.fluxmodel, FactorizedFluxModel):
                raise TypeError(
                    'The fluxmodel of the source hypothesis group must be an '
                    'instance of FactorizedFluxModel! '
                    f'Its current type is {classname(shg.fluxmodel)}!')

    def construct_detsigyield(
            self,
            dataset,
            data,
            shg,
            ppbar=None):
        """Constructs a detector signal yield 2-dimensional log spline
        function for the given flux model with varying parameter values.

        Parameters
        ----------
        dataset : instance of Dataset
            The Dataset instance holding the sin(dec) binning definition.
        data : instance of DatasetData
            The instance of DatasetData holding the monte-carlo event data.
            This implementation loads the effective area from the provided
            public data and hence does not need monte-carlo data.
        shg : instance of SourceHypoGroup
            The instance of SourceHypoGroup (i.e. sources and flux model) for
            which the detector signal yield should get constructed.
        ppbar : ProgressBar instance | None
            The instance of ProgressBar of the optional parent progress bar.

        Returns
        -------
        detsigyield : instance of SingleParamFluxPointLikeSourceI3DetSigYield
            The DetSigYield instance for a point-like source with a flux model
            of a single parameter.
        """
        self.assert_types_of_construct_detsigyield_arguments(
            dataset=dataset,
            data=data,
            shgs=shg,
            ppbar=ppbar,
        )

        # Get integrated live-time in days.
        livetime_days = Livetime.get_integrated_livetime(data.livetime)

        to_internal_time_unit_factor = self._cfg.to_internal_time_unit(
            time_unit=units.day
        )

        # Calculate conversion factor from the flux model unit into the internal
        # flux unit GeV^-1 cm^-2 s^-1.
        to_internal_flux_unit_factor = shg.fluxmodel.to_internal_flux_unit()

        # Load the effective area data from the public dataset.
        aeff_fnames = dataset.get_abs_pathfilename_list(
            dataset.get_aux_data_definition('eff_area_datafile'))
        (
            aeff_arr,
            sin_true_dec_binedges_lower,
            sin_true_dec_binedges_upper,
            log_true_e_binedges_lower,
            log_true_e_binedges_upper
        ) = load_effective_area_array(aeff_fnames)

        # Calculate the detector signal yield in sin_dec vs gamma.
        def _create_hist(
                energy_bin_edges_lower,
                energy_bin_edges_upper,
                aeff,
                fluxmodel,
                to_internal_flux_unit_factor,
        ):
            """Creates a histogram of the detector signal yield for the given
            sin(dec) binning.

            Parameters
            ----------
            energy_bin_edges_lower : 1d ndarray
                The array holding the lower bin edges in E_nu/GeV.
            energy_bin_edges_upper : 1d ndarray
                The array holding the upper bin edges in E_nu/GeV.
            aeff : (n_bins_sin_dec, n_bins_log_energy)-shaped 2d ndarray
                The effective area binned data array.

            Returns
            -------
            h : instance of ndarray
                The (n_bins_sin_dec,)-shaped 1d numpy ndarray containing the
                detector signal yield values for the different sin_dec bins and
                the given flux model.
            """
            h_phi = fluxmodel.energy_profile.get_integral(
                E1=energy_bin_edges_lower,
                E2=energy_bin_edges_upper)
            h_phi *= to_internal_flux_unit_factor

            # Sum over the enegry bins for each sin_dec row.
            h = np.sum(aeff*h_phi, axis=1)

            return h

        energy_bin_edges_lower = np.power(10, log_true_e_binedges_lower)
        energy_bin_edges_upper = np.power(10, log_true_e_binedges_upper)

        # Make a copy of the parameter grid and extend the grid by one bin on
        # each side.
        param_grid = self.param_grid.copy()
        param_grid.add_extra_lower_and_upper_bin()

        # Construct the arguments for the hist function to be used in the
        # multiproc.parallelize function.
        args_list = [
            (
                (
                    energy_bin_edges_lower,
                    energy_bin_edges_upper,
                    aeff_arr,
                    shg.fluxmodel.copy({param_grid.name: param_val}),
                    to_internal_flux_unit_factor,
                ),
                {}
            )
            for param_val in param_grid.grid
        ]
        h = np.vstack(
            multiproc.parallelize(
                _create_hist, args_list, self.ncpu, ppbar=ppbar)).T
        h *= livetime_days*to_internal_time_unit_factor

        # Create a 2d spline in log of the detector signal yield.
        sin_dec_bincenters = 0.5*(
            sin_true_dec_binedges_lower + sin_true_dec_binedges_upper)
        log_spl_sinDec_param = scipy.interpolate.RectBivariateSpline(
            sin_dec_bincenters,
            param_grid.grid,
            np.log(h),
            kx=self.spline_order_sinDec,
            ky=self.spline_order_param,
            s=0)

        # Construct the detector signal yield instance with the created spline.
        sin_dec_binedges = np.concatenate(
            (sin_true_dec_binedges_lower, [sin_true_dec_binedges_upper[-1]]))
        sin_dec_binning = BinningDefinition('sin_dec', sin_dec_binedges)

        detsigyield = SingleParamFluxPointLikeSourceI3DetSigYield(
            param_name=param_grid.name,
            dataset=dataset,
            fluxmodel=shg.fluxmodel,
            livetime=data.livetime,
            sin_dec_binning=sin_dec_binning,
            log_spl_sinDec_param=log_spl_sinDec_param)

        return detsigyield



# -*- coding: utf-8 -*-

"""The timing module provides code execution timing functionalities. The
TimeLord class keeps track of execution times of specific code segments,
called "tasks". The TaskTimer class can be used within a `with`
statement to time the execution of the code within the `with` block.
"""

import numpy as np
import time

from skyllh.core import (
    display,
)
from skyllh.core.py import (
    classname,
)


class TaskRecord(
        object):
    def __init__(
            self,
            name,
            start_times,
            end_times):
        """Creates a new TaskRecord instance.

        Parameters
        ----------
        name : str
            The name of the task.
        start_times : list of float
            The start times of the task in seconds.
        end_times : list of float
            The end times of the task in seconds.
        """
        self.name = name

        if len(start_times) != len(end_times):
            raise ValueError(
                'The number of start and end time stamps must be equal!')

        self._start_times = start_times
        self._end_times = end_times

    @property
    def tstart(self):
        """(read-only) The time stamps the execution of this task started.
        """
        return self._start_times

    @property
    def tend(self):
        """(read-only) The time stamps the execution of this task was stopped.
        """
        return self._end_times

    @property
    def duration(self):
        """(read-only) The total duration (without time overlap) the task was
        executed.
        """
        # Create a (2,Niter)-shaped 2D ndarray holding the start and end time
        # stamps of the task executions. This array gets then sorted by the
        # start time stamps.
        arr = np.sort(
            np.vstack((self._start_times, self._end_times)),
            axis=1)

        d = arr[1, 0] - arr[0, 0]
        last_tend = arr[1, 0]
        for idx in range(1, arr.shape[1]):
            tstart = arr[0, idx]
            tend = arr[1, idx]
            if tend <= last_tend:
                continue
            if tstart <= last_tend and tend > last_tend:
                d += tend - last_tend
            elif tstart >= last_tend:
                d += tend - tstart
            last_tend = tend

        return d

    @property
    def niter(self):
        """(read-only) The number of times this task was executed.
        """
        return len(self._start_times)

    def join(self, tr):
        """Joins this TaskRecord with the given TaskRecord instance.

        Parameters
        ----------
        tr : instance of TaskRecord
            The instance of TaskRecord that should be joined with this
            TaskRecord instance.
        """
        self._start_times.extend(tr._start_times)
        self._end_times.extend(tr._end_times)


class TimeLord(
        object):
    def __init__(
            self):
        self._task_records = []
        self._task_records_name_idx_map = {}

    @property
    def task_name_list(self):
        """(read-only) The list of task names.
        """
        return list(self._task_records_name_idx_map.keys())

    def add_task_record(
            self,
            tr):
        """Adds a given task record to the internal list of task records.
        """
        tname = tr.name

        if self.has_task_record(tname):
            # The TaskRecord already exists. Update the task record.
            self_tr = self.get_task_record(tname)
            self_tr.join(tr)
            return

        self._task_records.append(tr)
        self._task_records_name_idx_map[tr.name] = len(self._task_records)-1

    def get_task_record(
            self,
            name):
        """Retrieves a task record of the given name.

        Parameters
        ----------
        name : str
            The name of the task record.

        Returns
        -------
        task_record : instance of TaskRecord
            The instance of TaskRecord with the requested name.
        """
        return self._task_records[self._task_records_name_idx_map[name]]

    def has_task_record(
            self,
            name):
        """Checks if this TimeLord instance has a task record of the given name.

        Parameters
        ----------
        name : str
            The name of the task record.

        Returns
        -------
        check : bool
            ``True`` if this TimeLord instance has a task record of the given
            name, and ``False`` otherwise.
        """
        return name in self._task_records_name_idx_map

    def join(
            self,
            tl):
        """Joins a given TimeLord instance with this TimeLord instance. Tasks
        of the same name will be updated and new tasks will be added.

        Parameters
        ----------
        tl : instance of TimeLord
            The instance of TimeLord whos tasks should be joined with the tasks
            of this TimeLord instance.
        """
        for tname in tl.task_name_list:
            other_tr = tl.get_task_record(tname)
            if self.has_task_record(tname):
                # Update the task record.
                tr = self.get_task_record(tname)
                tr.join(other_tr)
            else:
                # Add a new task record.
                self.add_task_record(other_tr)

    def task_timer(self, name):
        """Creates TaskTimer instance for the given task name.
        """
        return TaskTimer(self, name)

    def __str__(self):
        """Generates a pretty string for this time lord.
        """
        task_name_list = self.task_name_list
        n_tasks = len(task_name_list)

        s = f'{classname(self)}: Executed tasks:'
        if n_tasks == 0:
            s += ' None.'
            return s

        task_name_len_list = [
            len(task_name)
            for task_name in task_name_list
        ]
        max_task_name_len = np.minimum(
            np.max(task_name_len_list), display.PAGE_WIDTH-25)

        for i in range(n_tasks):
            tr = self._task_records[i]
            task_name = tr.name[0:max_task_name_len]
            t = tr.duration / tr.niter
            line = '\n[{task_name:'+str(max_task_name_len)+'s}] {t:7.{p}{c}} '\
                'sec/iter ({niter:d})'
            s += line.format(
                task_name=task_name,
                t=t,
                p=1 if t > 1e3 or t < 1e-3 else 3,
                c='e' if t > 1e3 or t < 1e-3 else 'f',
                niter=tr.niter)

        return s


class TaskTimer(
        object):
    def __init__(
            self,
            time_lord,
            name):
        """
        Parameters
        ----------
        time_lord : instance of TimeLord
            The TimeLord instance that keeps track of the recorded tasks.
        name : str
            The name of the task.
        """
        self.time_lord = time_lord
        self.name = name

        self._start = None
        self._end = None

    @property
    def time_lord(self):
        """The TimeLord instance that keeps track of the recorded tasks. This
        can be None, which means that the task should not get recorded.
        """
        return self._time_lord

    @time_lord.setter
    def time_lord(self, lord):
        if lord is not None:
            if not isinstance(lord, TimeLord):
                raise TypeError(
                    'The time_lord property must be None or an instance of '
                    'TimeLord!')
        self._time_lord = lord

    @property
    def name(self):
        """The name if the task.
        """
        return self._name

    @name.setter
    def name(self, name):
        if not isinstance(name, str):
            raise TypeError(
                'The name property must be an instance of str!')
        self._name = name

    @property
    def duration(self):
        """The duration in seconds the task was executed.
        """
        return (self._end - self._start)

    def __enter__(self):
        """This gets executed when entering the `with` block.
        """
        self._start = time.process_time()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """This gets executed when exiting the `with` block.
        """
        self._end = time.process_time()

        if self._time_lord is None:
            return

        self._time_lord.add_task_record(
            TaskRecord(
                name=self._name,
                start_times=[self._start],
                end_times=[self._end]))



# -*- coding: utf-8 -*-

"""The analysis module provides classes for pre-defined analyses.
"""

import abc
from astropy import units
import numpy as np


from skyllh.core.background_generator import (
    BackgroundGenerator,
    MultiDatasetBackgroundGenerator,
)
from skyllh.core.config import (
    HasConfig,
)
from skyllh.core.dataset import (
    Dataset,
    DatasetData,
)
from skyllh.core.debugging import (
    get_logger,
)
from skyllh.core.event_selection import (
    EventSelectionMethod,
)
from skyllh.core.llhratio import (
    LLHRatio,
    MultiDatasetTCLLHRatio,
    ZeroSigH0SingleDatasetTCLLHRatio,
)
from skyllh.core.multiproc import (
    get_ncpu,
    parallelize,
)
from skyllh.core.parameters import (
    ParameterModelMapper,
)
from skyllh.core.pdfratio import (
    PDFRatio,
    SourceWeightedPDFRatio,
)
from skyllh.core.py import (
    classname,
    issequenceof,
)
from skyllh.core.random import (
    RandomStateService,
)
from skyllh.core.services import (
    DatasetSignalWeightFactorsService,
    DetSigYieldService,
    SrcDetSigYieldWeightsService,
)
from skyllh.core.signal_generator import (
    SignalGenerator,
    MultiDatasetSignalGenerator,
)
from skyllh.core.source_hypo_grouping import (
    SourceHypoGroupManager,
)
from skyllh.core.source_model import (
    SourceModel,
)
from skyllh.core.storage import (
    DataFieldRecordArray,
)
from skyllh.core.test_statistic import (
    TestStatistic,
)
from skyllh.core.timing import (
    TaskTimer,
)
from skyllh.core.trialdata import (
    TrialDataManager,
)


logger = get_logger(__name__)


class Analysis(
        HasConfig,
        metaclass=abc.ABCMeta,
):
    """This is the abstract base class for all analysis classes.
    It contains common properties required by all analyses and defines the
    overall analysis interface how to setup and run an analysis.
    """
    def __init__(
            self,
            shg_mgr,
            pmm,
            test_statistic,
            bkg_generator_cls=None,
            sig_generator_cls=None,
            **kwargs):
        """Constructor of the analysis base class.

        Parameters
        ----------
        shg_mgr : instance of SourceHypoGroupManager
            The instance of SourceHypoGroupManager, which defines the groups of
            source hypotheses, their flux model, and their detector signal
            yield implementation method.
        pmm : instance of ParameterModelMapper
            The ParameterModelMapper instance managing the global set of
            parameters and their relation to individual models, e.g. sources.
        test_statistic : TestStatistic instance
            The TestStatistic instance that defines the test statistic function
            of the analysis.
        bkg_generator_cls : class of MultiDatasetBackgroundGenerator | None
            The background generator class used to create the background
            generator instance for multiple datasets.
            If set to ``None``, the
            :class:`skyllh.core.background_generator.MultiDatasetBackgroundGenerator`
            class is used.
        sig_generator_cls : class of MultiDatasetSignalGenerator | None
            The signal generator class used to create the signal generator
            instance for multiple datasets.
            If set to ``None``, the
            :class:`~skyllh.core.signal_generator.MultiDatasetSignalGenerator`
            class is used.
        """
        super().__init__(
            **kwargs)

        self.shg_mgr = shg_mgr
        self.pmm = pmm
        self.test_statistic = test_statistic
        self.bkg_generator_cls = bkg_generator_cls
        self.sig_generator_cls = sig_generator_cls

        self._dataset_list = []
        self._data_list = []
        self._tdm_list = []
        self._event_selection_method_list = []

        self._detsigyield_service = None
        self._src_detsigyield_weights_service = None
        self._ds_sig_weight_factors_service = None

        self._bkg_generator_list = []
        self._bkg_generator = None
        self._sig_generator_list = []
        self._sig_generator = None

    @property
    def shg_mgr(self):
        """The SourceHypoGroupManager instance, which defines the groups of
        source hypothesis, their flux model, and their detector signal
        yield implementation method.
        """
        return self._shg_mgr

    @shg_mgr.setter
    def shg_mgr(self, mgr):
        if not isinstance(mgr, SourceHypoGroupManager):
            raise TypeError(
                'The shg_mgr property must be an instance of '
                'SourceHypoGroupManager! '
                f'Its current type is {classname(mgr)}.')
        self._shg_mgr = mgr

    @property
    def pmm(self):
        """The ParameterModelMapper instance that manages the global set of
        parameters and their relation to individual models, e.g. sources.
        """
        return self._pmm

    @pmm.setter
    def pmm(self, mapper):
        if not isinstance(mapper, ParameterModelMapper):
            raise TypeError(
                'The pmm property must be an instance of '
                'ParameterModelMapper! '
                f'Its current type is {classname(mapper)}.')
        self._pmm = mapper

    @property
    def test_statistic(self):
        """The TestStatistic instance that defines the test-statistic function
        of the analysis.
        """
        return self._test_statistic

    @test_statistic.setter
    def test_statistic(self, ts):
        if not isinstance(ts, TestStatistic):
            raise TypeError(
                'The test_statistic property must be an instance of '
                'TestStatistic! '
                f'Its current type is {classname(ts)}.')
        self._test_statistic = ts

    @property
    def bkg_generator_list(self):
        """(read-only) The list of instance of BackgroundGenerator, one for each
        dataset.
        """
        return self._bkg_generator_list

    @property
    def bkg_generator_cls(self):
        """The background generator class that should be used to construct the
        background generator instance.
        """
        return self._bkg_generator_cls

    @bkg_generator_cls.setter
    def bkg_generator_cls(self, cls):
        if cls is None:
            cls = MultiDatasetBackgroundGenerator
        if not issubclass(cls, BackgroundGenerator):
            raise TypeError(
                'The bkg_generator_cls property must be a subclass of '
                'BackgroundGenerator! '
                f'Its current type is {classname(cls)}.')
        self._bkg_generator_cls = cls

    @property
    def dataset_list(self):
        """The list of Dataset instances.
        """
        return self._dataset_list

    @dataset_list.setter
    def dataset_list(self, datasets):
        if not issequenceof(datasets, Dataset):
            raise TypeError(
                'The dataset_list property must be a sequence of Dataset '
                'instances! '
                f'Its current type is {classname(datasets)}.')
        self._dataset_list = list(datasets)

    @property
    def data_list(self):
        """The list of DatasetData instances holding the original data of the
        dataset.
        """
        return self._data_list

    @data_list.setter
    def data_list(self, datas):
        if not issequenceof(datas, DatasetData):
            raise TypeError(
                'The data_list property must be a sequence of DatasetData '
                'instances! '
                f'Its current type is {classname(datas)}.')
        self._data_list = list(datas)

    @property
    def n_datasets(self):
        """(read-only) The number of datasets used in this analysis.
        """
        return len(self._dataset_list)

    @property
    def bkg_generator(self):
        """(read-only) The background generator instance. Is None of the
        background generator has not been constructed via the
        `construct_background_generator` method.
        """
        return self._bkg_generator

    @property
    def detsigyield_service(self):
        """The instance of DetSigYieldService for the analysis.
        """
        return self._detsigyield_service

    @detsigyield_service.setter
    def detsigyield_service(self, service):
        if not isinstance(service, DetSigYieldService):
            raise TypeError(
                'The detsigyield_service property must be an instance of '
                'DetSigYieldService! '
                f'Its current type is {classname(service)}!')
        self._detsigyield_service = service

    @property
    def src_detsigyield_weights_service(self):
        """The instance of SrcDetSigYieldWeightsService for the analysis.
        """
        return self._src_detsigyield_weights_service

    @src_detsigyield_weights_service.setter
    def src_detsigyield_weights_service(self, service):
        if not isinstance(service, SrcDetSigYieldWeightsService):
            raise TypeError(
                'The src_detsigyield_weights_service property must be an '
                'instance of SrcDetSigYieldWeightsService! '
                f'Its current type is {classname(service)}!')
        self._src_detsigyield_weights_service = service

    @property
    def ds_sig_weight_factors_service(self):
        """The instance of DatasetSignalWeightFactorsService for the analysis.
        """
        return self._ds_sig_weight_factors_service

    @ds_sig_weight_factors_service.setter
    def ds_sig_weight_factors_service(self, service):
        if not isinstance(service, DatasetSignalWeightFactorsService):
            raise TypeError(
                'The ds_sig_weight_factors_service property must be an '
                'instance of DatasetSignalWeightFactorsService! '
                f'Its current type is {classname(service)}!')
        self._ds_sig_weight_factors_service = service

    @property
    def sig_generator_list(self):
        """(read-only) The list of instance of SignalGenerator, one for each
        dataset.
        """
        return self._sig_generator_list

    @property
    def sig_generator_cls(self):
        """The signal generator class that should be used to construct the
        signal generator instance handling all datasets.
        """
        return self._sig_generator_cls

    @sig_generator_cls.setter
    def sig_generator_cls(self, cls):
        if cls is None:
            cls = MultiDatasetSignalGenerator
        if not issubclass(cls, SignalGenerator):
            raise TypeError(
                'The sig_generator_cls property must be a subclass of '
                'SignalGenerator! '
                f'Its current type is {classname(cls)}.')
        self._sig_generator_cls = cls

    @property
    def sig_generator(self):
        """(read-only) The signal generator instance. Is None if the signal
        generator has not been constructed via the
        `construct_signal_generator` method.
        """
        return self._sig_generator

    @property
    def tdm_list(self):
        """The list of instance of TrialDataManager. One for each dataset.
        """
        return self._tdm_list

    @tdm_list.setter
    def tdm_list(self, tdms):
        if not issequenceof(tdms, TrialDataManager):
            raise TypeError(
                'The tdm_list property must be a sequence of TrialDataManager '
                'instances! '
                f'Its current type is {classname(tdms)}.')
        self._tdm_list = list(tdms)

    @property
    def total_livetime(self):
        """(read-only) The total live-time in days of the loaded data.
        """
        livetime = 0
        for data in self._data_list:
            livetime += data.livetime
        return livetime

    def construct_services(
            self,
            ppbar=None,
    ):
        """Constructs the following services:

            - detector signal yield service
            - source detector signal yield weights service
            - dataset signal weight factors service

        Parameters
        ----------
        ppbar : instance of ProgressBar | None
            The instance of ProgressBar of the optional parent progress bar.
        """
        self.detsigyield_service = DetSigYieldService(
            shg_mgr=self._shg_mgr,
            dataset_list=self._dataset_list,
            data_list=self._data_list,
            ppbar=ppbar,
        )

        self.src_detsigyield_weights_service = SrcDetSigYieldWeightsService(
            detsigyield_service=self.detsigyield_service,
        )

        self.ds_sig_weight_factors_service = DatasetSignalWeightFactorsService(
            src_detsigyield_weights_service=self.src_detsigyield_weights_service,
        )

    def add_dataset(  # noqa: C901
            self,
            dataset,
            data,
            tdm=None,
            event_selection_method=None,
            bkg_generator=None,
            sig_generator=None,
    ):
        """Adds the given dataset to the list of datasets for this analysis.

        Parameters
        ----------
        dataset : instance of Dataset
            The Dataset instance that should get added.
        data : instance of DatasetData
            The DatasetData instance holding the original (prepared) data of the
            dataset.
        tdm : instance of TrialDataManager | None
            The TrialDataManager instance managing the trial data and additional
            data fields of the data set. If set to None, it means that no
            additional data fields are defined.
        event_selection_method : instance of EventSelectionMethod | None
            The instance of EventSelectionMethod to use to select only
            signal-like events from the data. All other events
            will be treated as pure background events. This reduces the amount
            of log-likelihood-ratio function evaluations. If set to None, all
            events will be evaluated.
        bkg_generator : instance of BackgroundGenerator | None
            The optional instance of BackgroundGenerator, which should be used
            to generate background events for this particular dataset.
        sig_generator : instance of SignalGenerator | None
            The optional instance of SignalGenerator, which should be used
            to generate signal events for this particular dataset.
        """
        if not isinstance(dataset, Dataset):
            raise TypeError(
                'The dataset argument must be an instance of Dataset!')

        if not isinstance(data, DatasetData):
            raise TypeError(
                'The data argument must be an instance of DatasetData!')

        if tdm is None:
            tdm = TrialDataManager()
        if not isinstance(tdm, TrialDataManager):
            raise TypeError(
                'The tdm argument must be None or an instance of '
                'TrialDataManager! '
                f'Its current type is {classname(tdm)}!')

        if event_selection_method is not None:
            if not isinstance(event_selection_method, EventSelectionMethod):
                raise TypeError(
                    'The event_selection_method argument must be None or an '
                    'instance of EventSelectionMethod! '
                    f'Its current type is {classname(event_selection_method)}!')

        if bkg_generator is not None:
            if not isinstance(bkg_generator, BackgroundGenerator):
                raise TypeError(
                    'The bkg_generator argument must be None or an instance of'
                    'BackgroundGenerator! '
                    f'Its current type is {classname(bkg_generator)}!')

        if sig_generator is not None:
            if not isinstance(sig_generator, SignalGenerator):
                raise TypeError(
                    'The sig_generator argument must be None or an instance of '
                    'SignalGenerator! '
                    f'Its current type is {classname(sig_generator)}!')

        self._dataset_list.append(dataset)
        self._data_list.append(data)
        self._tdm_list.append(tdm)
        self._event_selection_method_list.append(event_selection_method)
        self._bkg_generator_list.append(bkg_generator)
        self._sig_generator_list.append(sig_generator)

    def get_livetime(
            self,
            dataset_key=None,
            unit=None):
        """Retrieves the numeric livetime of the given dataset in the specified
        unit. The dataset can be specified either through its index or its name.
        If no dataset is specified, the total livetime, i.e. the sum of the
        livetime of all datasets, is returned.

        Parameters
        ----------
        dataset_key : int | str | None
            The index or name of the dataset for which the livetime should get
            retrieved. If set to ``None``, the total livetime of all datasets
            will be returned.
        unit : instance of astropy.units.Unit | None
            The time unit in which the livetime should be returned. If set to
            ``None``, ``astropy.units.day`` will be used.
        """
        if dataset_key is None:
            livetime = np.sum([data.livetime for data in self._data_list])
        else:
            if isinstance(dataset_key, int):
                dataset_idx = dataset_key
            elif isinstance(dataset_key, str):
                dataset_idx = None
                for (idx, ds) in enumerate(self._dataset_list):
                    if ds.name == dataset_key:
                        dataset_idx = idx
                        break
                if dataset_idx is None:
                    raise KeyError(
                        f'The dataset of name "{dataset_key}" does not exist!')
            else:
                raise TypeError(
                    'The dataset_key argument must be an instance of int, str, '
                    'or None! '
                    f'Its current type is {classname(dataset_key)}.')
            livetime = self._data_list[dataset_idx].livetime

        if isinstance(unit, units.Unit):
            livetime *= units.day.to(unit)

        return livetime

    def calculate_test_statistic(
            self,
            log_lambda,
            fitparam_values,
            **kwargs):
        """Calculates the test statistic value by calling the ``evaluate``
        method of the TestStatistic class with the given log_lambda value and
        fit parameter values.

        Parameters
        ----------
        log_lambda : float
            The value of the log-likelihood ratio function. Usually, this is its
            maximum.
        fitparam_values : instance of numpy ndarray
            The (N_fitparam,)-shaped 1D ndarray holding the global
            fit parameter values of the log-likelihood ratio function for
            the given log_lambda value.
        **kwargs
            Any additional keyword arguments are passed to the
            ``__call__`` method of the TestStatistic instance.

        Returns
        -------
        TS : float
            The calculated test-statistic value.
        """
        return self._test_statistic(
            pmm=self._pmm,
            log_lambda=log_lambda,
            fitparam_values=fitparam_values,
            **kwargs)

    def construct_background_generator(self, **kwargs):
        """Constructs the background generator for all added datasets.
        This method must be called after all the datasets were added via the
        add_dataset method. It sets the `bkg_generator` property of this
        Analysis class instance.
        """
        self._bkg_generator = self.bkg_generator_cls(
            cfg=self._cfg,
            dataset_list=self._dataset_list,
            data_list=self._data_list,
            bkg_generator_list=self._bkg_generator_list,
            **kwargs)

    def construct_signal_generator(self, **kwargs):
        """Constructs the signal generator for all added datasets.
        This method must be called after all the datasets were added via the
        add_dataset method. It sets the `sig_generator` property of this
        Analysis class instance. The signal generation method has to be set
        through the source hypothesis group.
        """
        self._sig_generator = self.sig_generator_cls(
            cfg=self._cfg,
            shg_mgr=self._shg_mgr,
            dataset_list=self._dataset_list,
            data_list=self._data_list,
            sig_generator_list=self._sig_generator_list,
            ds_sig_weight_factors_service=self.ds_sig_weight_factors_service,
            **kwargs)

    @abc.abstractmethod
    def initialize_trial(
            self,
            events_list,
            n_events_list=None):
        """This method is supposed to initialize the log-likelihood ratio
        function with a new set of given trial data. This is a low-level method.
        For convenient methods see the `unblind` and `do_trial` methods.

        Parameters
        ----------
        events_list : list of numpy record ndarray
            The list of data events to use for the log-likelihood function
            evaluation. The data arrays for the datasets must be in the same
            order than the added datasets.
        n_events_list : list of int | None
            The list of the number of events of each data set. These numbers
            can be larger than the number of events given by the `events_list`
            argument in cases where an event selection method was already used.
            If set to None, the number of events is taken from the given
            `events_list` argument.
        """
        pass

    @abc.abstractmethod
    def unblind(
            self,
            minimizer_rss,
            tl=None):
        """This method is supposed to run the analysis on the experimental data,
        i.e. unblinds the data.

        Parameters
        ----------
        minimizer_rss : instance of RandomStateService
            The instance of RandomStateService that should be used by the
            minimizer to generate new random initial fit parameter values.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to time the
            maximization of the LLH ratio function.

        Returns
        -------
        TS : float
            The test-statistic value.
        global_params_dict : dict
            The dictionary holding the global parameter names and their
            best fit values. It includes fixed and floating parameters.
        status : dict
            The status dictionary with information about the performed
            minimization process of the analysis.
        """
        pass

    @abc.abstractmethod
    def do_trial_with_given_pseudo_data(
            self,
            seed,
            mean_n_sig,
            n_sig,
            n_events_list,
            events_list,
            minimizer_rss,
            minimizer_status_dict=None,
            tl=None,
            **kwargs,
    ):
        """This method is supposed to perform an analysis trial on a given
        pseudo data.

        Parameters
        ----------
        seed : int
            The seed value, which was used to generate the pseudo data. It will
            be stored in the returned result array.
        mean_n_sig : float
            The mean number of signal events the pseudo data was generated with.
        n_sig : int
            The total number of actual signal events in the pseudo data.
        n_events_list : list of int
            The total number of events for each data set of the pseudo data.
        events_list : list of instance of DataFieldRecordArray
            The list of instance of DataFieldRecordArray containing the pseudo
            data events for each data sample. The number of events for each
            data sample can be less than the number of events given by
            ``n_events_list`` if an event selection method was already utilized
            when generating background events.
        minimizer_rss : instance of RandomStateService
            The instance of RandomStateService to use for generating random
            numbers for the minimizer, e.g. for new initial fit parameter
            values.
        minimizer_status_dict : dict | None
            If a dictionary is provided, it will be updated with the minimizer
            status dictionary.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to time
            individual tasks.

        Returns
        -------
        recarray : instance of numpy record ndarray
            The numpy record ndarray holding the result of the trial. It must
            contain the following data fields:

            rss_seed : int
                The RandomStateService seed.
            mean_n_sig : float
                The mean number of signal events.
            n_sig : int
                The actual number of injected signal events.
            ts : float
                The test-statistic value.
            [<global_param_name> : float ]
                Any additional parameters of the analysis.
        """
        pass

    def change_shg_mgr(
            self,
            shg_mgr,
            update_detsigyield_service=True):
        """If the SourceHypoGroupManager instance changed, this method needs to
        be called to propagate the change to all components of the analysis.

        Parameters
        ----------
        shg_mgr : instance of SourceHypoGroupManager
            The new instance of SourceHypoGroupManager.
        update_detsigyield_service : bool
            The option whether to update SourceHypoGroupManager of
            `detsigyield_service` property. It can be set to false for runtime
            optimization when the detector signal yield does not change for the
            new source. Default is True.
        """
        for evt_selection_method in self._event_selection_method_list:
            if evt_selection_method is not None:
                evt_selection_method.change_shg_mgr(
                    shg_mgr=shg_mgr)

        for tdm in self._tdm_list:
            tdm.change_shg_mgr(
                shg_mgr=shg_mgr,
                pmm=self._pmm)

        if self._detsigyield_service is not None and update_detsigyield_service:
            self._detsigyield_service.change_shg_mgr(
                shg_mgr=shg_mgr)

        if self._src_detsigyield_weights_service is not None:
            self._src_detsigyield_weights_service.change_shg_mgr(
                shg_mgr=shg_mgr)

        if self._bkg_generator is not None:
            self._bkg_generator.change_shg_mgr(
                shg_mgr=shg_mgr)

        if self._sig_generator is not None:
            self._sig_generator.change_shg_mgr(
                shg_mgr=shg_mgr)

    def do_trial_with_given_bkg_and_sig_pseudo_data(
            self,
            seed,
            mean_n_sig,
            n_sig,
            n_bkg_events_list,
            n_sig_events_list,
            bkg_events_list,
            sig_events_list,
            minimizer_rss,
            minimizer_status_dict=None,
            tl=None,
            **kwargs,
    ):
        """Performs an analysis trial on the given background and signal pseudo
        data. This method merges the background and signal pseudo events and
        calls the ``do_trial_with_given_pseudo_data`` method of this class.

        Note
        ----
        This method alters the DataFieldRecordArray instances of the
        bkg_events_list argument!

        Parameters
        ----------
        seed : int
            The seed value, which was used to generate the background and signal
            pseudo data. It will be stored in the returned result array.
        mean_n_sig : float
            The mean number of signal events the pseudo data was generated with.
        n_sig : int
            The total number of actual signal events in the pseudo data.
        n_bkg_events_list : list of int
            The total number of background events for each data set of the
            pseudo data.
        n_sig_events_list : list of int
            The total number of signal events for each data set of the
            pseudo data.
        bkg_events_list : list of instance of DataFieldRecordArray
            The list of instance of DataFieldRecordArray containing the
            background pseudo data events for each data set.
        sig_events_list : list of instance of DataFieldRecordArray | None
            The list of instance of DataFieldRecordArray containing the signal
            pseudo data events for each data set. If a particular dataset has
            no signal events, the entry for that dataset can be ``None``.
        minimizer_rss : instance of RandomStateService
            The instance of RandomStateService to use for generating random
            numbers for the minimizer, e.g. for new initial fit parameter
            values.
        minimizer_status_dict : dict | None
            If a dictionary is provided, it will be updated with the minimizer
            status dictionary.
        tl : instance of TimeLord | None
            The instance of TimeLord that should be used to time individual
            tasks.
        **kwargs : dict
            Additional keyword arguments are passed to the
            :meth:`~skyllh.core.analysis.Analysis.do_trial_with_given_pseudo_data`
            method.

        Returns
        -------
        recarray : instance of numpy record ndarray
            The numpy record ndarray holding the result of the trial.
            See the documentation of the
            :meth:`~skyllh.core.analysis.Analysis.do_trial_with_given_pseudo_data`
            method for further information.
        """
        n_events_list = list(
            np.array(n_bkg_events_list) +
            np.array(n_sig_events_list)
        )

        events_list = bkg_events_list

        # Add potential signal events to the background events.
        for ds_idx in range(len(events_list)):
            if sig_events_list[ds_idx] is not None:
                if events_list[ds_idx] is None:
                    events_list[ds_idx] = sig_events_list[ds_idx]
                else:
                    events_list[ds_idx].append(sig_events_list[ds_idx])

        recarray = self.do_trial_with_given_pseudo_data(
            seed=seed,
            mean_n_sig=mean_n_sig,
            n_sig=n_sig,
            n_events_list=n_events_list,
            events_list=events_list,
            minimizer_rss=minimizer_rss,
            minimizer_status_dict=minimizer_status_dict,
            tl=tl,
            **kwargs)

        return recarray

    def generate_background_events(
            self,
            rss,
            mean_n_bkg_list=None,
            bkg_kwargs=None,
            tl=None,
    ):
        """Generates background events utilizing the background generator.

        Parameters
        ----------
        rss : instance of RandomStateService
            The instance of RandomStateService to use for generating random
            numbers.
        mean_n_bkg_list : list of float | None
            The mean number of background events that should be generated for
            each dataset. If set to None (the default), the background
            generation method needs to obtain this number itself.
        bkg_kwargs : dict | None
            Optional keyword arguments for the ``generate_background_events``
            method of the background generator.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to time
            individual tasks of this method.

        Returns
        -------
        n_events_list : list of int
            The list of the number of events that have been generated for each
            pseudo data set.
        events_list : list of instance of DataFieldRecordArray
            The list of instance of DataFieldRecordArray containing the pseudo
            data events for each data sample. The number of events for each
            data set can be less than the number of events given by
            ``n_events_list`` if an event selection method was already utilized
            when generating background events.
        """
        if bkg_kwargs is None:
            bkg_kwargs = dict()

        if self._bkg_generator is None:
            self.construct_background_generator()

        (n_events_list, events_list) =\
            self._bkg_generator.generate_background_events(
                rss=rss,
                mean_n_bkg_list=mean_n_bkg_list,
                tl=tl,
                **bkg_kwargs,
            )

        return (n_events_list, events_list)

    def _assert_input_arguments_of_generate_signal_events(
            self,
            rss,
            n_events_list,
            events_list):
        """Checks the input arguments of the ``generate_signal_events`` method
        for correct type and value.
        """
        n_datasets = self.n_datasets

        if not isinstance(rss, RandomStateService):
            raise TypeError(
                'The rss argument must be an instance of RandomStateService! '
                f'Its current type is {classname(rss)}.')

        if not issequenceof(n_events_list, int):
            raise TypeError(
                'The n_events_list argument must be a sequence of '
                'instances of type int! '
                f'Its current type is {classname(n_events_list)}.')
        if len(n_events_list) != n_datasets:
            raise ValueError(
                'The n_events_list argument must be a list of int of '
                f'length {n_datasets}! Currently it is of length '
                f'{len(n_events_list)}.')

        if not issequenceof(events_list, (type(None), DataFieldRecordArray)):
            raise TypeError(
                'The events_list argument must be a sequence of '
                'instances of type DataFieldRecordArray! '
                f'Its current type is {classname(events_list)}.')
        if len(events_list) != n_datasets:
            raise ValueError(
                'The events_list argument must be a list of instances of '
                f'type DataFieldRecordArray with a length of {n_datasets}! '
                f'Currently it is of length {len(events_list)}.')

    def generate_signal_events(
            self,
            rss,
            mean_n_sig,
            sig_kwargs=None,
            n_events_list=None,
            events_list=None,
            tl=None):
        """Generates signal events utilizing the signal generator.

        Parameters
        ----------
        rss : instance of RandomStateService
            The instance of RandomStateService to use for generating random
            numbers.
        mean_n_sig : float
            The mean number of signal events that should be generated for the
            trial. The actual number of generated events will be drawn from a
            Poisson distribution with this given signal mean as mean.
        sig_kwargs : dict | None
            Additional keyword arguments for the ``generate_signal_events``
            method of the ``sig_generator_cls`` class. An usual keyword argument
            is ``poisson``.
        n_events_list : list of int | None
            If given, it specifies the number of events of each data set already
            present and the number of signal events will be added.
        events_list : list of instance of DataFieldRecordArray | None
            If given, it specifies the events of each data set already present
            and the signal events will be added.
        tl : instance of TimeLord | None
            The instance of TimeLord that should be used to time individual
            tasks of this method.

        Returns
        -------
        n_sig : int
            The actual number of injected signal events.
        n_events_list : list of int
            The list of the number of signal events that have been generated for
            each data set.
        events_list : list of instance of DataFieldRecordArray
            The list of instance of DataFieldRecordArray containing the
            signal data events for each data set. An entry is None, if no signal
            events were generated for this particular data set.
        """
        if sig_kwargs is None:
            sig_kwargs = dict()

        if n_events_list is None:
            n_events_list = [0] * self.n_datasets

        if events_list is None:
            events_list = [None] * self.n_datasets

        self._assert_input_arguments_of_generate_signal_events(
            rss=rss,
            n_events_list=n_events_list,
            events_list=events_list)

        n_sig = 0

        if mean_n_sig == 0:
            return (n_sig, n_events_list, events_list)

        # Construct the signal generator if not done yet.
        if self._sig_generator is None:
            with TaskTimer(tl, 'Constructing signal generator.'):
                self.construct_signal_generator()

        # Generate signal events with the given mean number of signal
        # events.
        sig_kwargs.update(mean=mean_n_sig)
        with TaskTimer(tl, 'Generating signal events.'):
            (n_sig, ds_sig_events_dict) =\
                self._sig_generator.generate_signal_events(
                    rss=rss,
                    **sig_kwargs)

        # Inject the signal events to the generated background data.
        for (ds_idx, sig_events) in ds_sig_events_dict.items():
            n_events_list[ds_idx] += len(sig_events)
            if events_list[ds_idx] is None:
                events_list[ds_idx] = sig_events
            else:
                events_list[ds_idx].append(sig_events)

        return (n_sig, n_events_list, events_list)

    def generate_pseudo_data(
            self,
            rss,
            mean_n_bkg_list=None,
            mean_n_sig=0,
            bkg_kwargs=None,
            sig_kwargs=None,
            tl=None):
        """Generates pseudo data with background and possible signal
        events for each data set using the background and signal generation
        methods of the analysis.

        Parameters
        ----------
        rss : instance of RandomStateService
            The instance of RandomStateService to use for generating random
            numbers.
        mean_n_bkg_list : list of float | None
            The mean number of background events that should be generated for
            each dataset. If set to None (the default), the background
            generation method needs to obtain this number itself.
        mean_n_sig : float
            The mean number of signal events that should be generated for the
            trial. The actual number of generated events will be drawn from a
            Poisson distribution with this given signal mean as mean.
        bkg_kwargs : dict | None
            Additional keyword arguments for the `generate_events` method of the
            background generation method class. An usual keyword argument is
            `poisson`.
        sig_kwargs : dict | None
            Additional keyword arguments for the `generate_signal_events` method
            of the `SignalGenerator` class. An usual keyword argument is
            `poisson`.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to time
            individual tasks of this method.

        Returns
        -------
        n_sig : int
            The actual number of injected signal events.
        n_events_list : list of int
            The list of the number of events that have been generated for each
            pseudo data set.
        events_list : list of instance of DataFieldRecordArray
            The list of DataFieldRecordArray instances containing the pseudo
            data events for each data sample. The number of events for each
            data set can be less than the number of events given by
            `n_events_list` if an event selection method was already utilized
            when generating background events.
        """
        # Generate background events for each dataset.
        (n_events_list, events_list) = self.generate_background_events(
            rss=rss,
            mean_n_bkg_list=mean_n_bkg_list,
            bkg_kwargs=bkg_kwargs,
            tl=tl)

        # Generate signal events and add them to the already generated
        # background events.
        (n_sig, n_events_list, events_list) = self.generate_signal_events(
            rss=rss,
            mean_n_sig=mean_n_sig,
            sig_kwargs=sig_kwargs,
            n_events_list=n_events_list,
            events_list=events_list,
            tl=tl)

        return (n_sig, n_events_list, events_list)

    def do_trial(
            self,
            rss,
            mean_n_bkg_list=None,
            mean_n_sig=0,
            bkg_kwargs=None,
            sig_kwargs=None,
            minimizer_rss=None,
            minimizer_status_dict=None,
            tl=None,
            **kwargs):
        """This method performs an analysis trial by generating a
        pseudo data sample with background events and possible signal events
        via the :meth:`generate_pseudo_data` method, and performs the analysis
        on that random pseudo data sample by calling the
        :meth:`do_trial_with_given_pseudo_data` method.

        Parameters
        ----------
        rss : instance of RandomStateService
            The instance of RandomStateService to use for generating
            random numbers.
        mean_n_bkg_list : list of float | None
            The mean number of background events that should be generated for
            each dataset. If set to None (the default), the background
            generation method needs to obtain this number itself.
        mean_n_sig : float
            The mean number of signal events that should be generated for the
            trial.
        bkg_kwargs : dict | None
            Additional keyword arguments for the `generate_events` method of the
            background generation method class. An usual keyword argument is
            `poisson`.
        sig_kwargs : dict | None
            Additional keyword arguments for the `generate_signal_events` method
            of the `SignalGenerator` class. An usual keyword argument is
            `poisson`.
        minimizer_rss : instance of RandomStateService | None
            The instance of RandomStateService to use for generating random
            numbers for the minimizer, e.g. new initial fit parameter values.
            If set to ``None``, ``rss`` will be used for the minimizer.
        minimizer_status_dict : dict | None
            If a dictionary is provided, it will be updated with the minimizer
            status dictionary.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to time
            individual tasks.
        **kwargs : dict
            Additional keyword arguments are passed to the
            :meth:`do_trial_with_given_pseudo_data` method.

        Returns
        -------
        recarray : instance of numpy record ndarray
            The numpy record ndarray holding the result of the trial.
            See the documentation of the
            :py:meth:`~skyllh.core.analysis.Analysis.do_trial_with_given_pseudo_data`
            method for further information.
        """
        if minimizer_rss is None:
            minimizer_rss = rss

        with TaskTimer(tl, 'Generating pseudo data.'):
            (n_sig, n_events_list, events_list) = self.generate_pseudo_data(
                rss=rss,
                mean_n_bkg_list=mean_n_bkg_list,
                mean_n_sig=mean_n_sig,
                bkg_kwargs=bkg_kwargs,
                sig_kwargs=sig_kwargs,
                tl=tl)

        recarray = self.do_trial_with_given_pseudo_data(
            seed=rss.seed,
            mean_n_sig=mean_n_sig,
            n_sig=n_sig,
            n_events_list=n_events_list,
            events_list=events_list,
            minimizer_rss=minimizer_rss,
            minimizer_status_dict=minimizer_status_dict,
            tl=tl,
            **kwargs)

        return recarray

    def do_trials(
            self,
            rss,
            n,
            ncpu=None,
            tl=None,
            ppbar=None,
            **kwargs):
        """Executes the :meth:`do_trial` method ``n`` times with possible
        multi-processing.

        Parameters
        ----------
        rss : instance of RandomStateService
            The RandomStateService instance to use for generating random
            numbers.
        n : int
            Number of trials to generate using the `do_trial` method.
        ncpu : int | None
            The number of CPUs to use, i.e. the number of subprocesses to
            spawn. If set to None, the global setting will be used.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to time
            individual tasks.
        ppbar : instance of ProgressBar | None
            The possible parent ProgressBar instance.
        **kwargs
            Additional keyword arguments are passed to the :meth:`do_trial`
            method. See the documentation of that method for allowed keyword
            arguments.

        Returns
        -------
        recarray : numpy record ndarray
            The numpy record ndarray holding the result of all trials.
            See the documentation of the
            :py:meth:`~skyllh.core.analysis.Analysis.do_trial` method for the
            list of data fields.
        """
        ncpu = get_ncpu(
            cfg=self._cfg,
            local_ncpu=ncpu)

        args_list = [((), kwargs) for i in range(n)]
        result_list = parallelize(
            func=self.do_trial,
            args_list=args_list,
            ncpu=ncpu,
            rss=rss,
            tl=tl,
            ppbar=ppbar)

        recarray_dtype = result_list[0].dtype
        recarray = np.empty(n, dtype=recarray_dtype)
        recarray[:] = np.array(result_list)[:, 0]

        return recarray


class LLHRatioAnalysis(
        Analysis,
        metaclass=abc.ABCMeta):
    """This is the abstract base class for all log-likelihood ratio analysis
    classes. It requires a mathematical log-likelihood ratio function.

    To set-up and run an analysis the following procedure applies:

        1. Create an Analysis instance.
        2. Add the datasets and their PDF ratio instances via the
           :meth:`skyllh.core.analysis.Analysis.add_dataset` method.
        3. Construct the log-likelihood ratio function via the
           :meth:`construct_llhratio` method.
        4. Initialize a trial via the :meth:`initialize_trial` method.
        5. Fit the global fit parameters to the trial data via the
           :meth:`maximize` method of the ``llhratio`` property.

    Alternatively, one can use the convenient methods :meth:`do_trial` or
    :meth:`unblind` to perform a random trial or to unblind the data,
    respectively. Both methods will fit the global fit parameters using the set
    up data. Finally, the test statistic is calculated via the
    :meth:`calculate_test_statistic` method.

    In order to calculate sensitivities and discovery potentials, analysis
    trials have to be performed on random data samples with injected signal
    events. To perform a trial with injected signal events, the signal generator
    has to be constructed via the :meth:`construct_signal_generator` method
    before any random trial data can be generated.
    """

    def __init__(
            self,
            shg_mgr,
            pmm,
            test_statistic,
            bkg_generator_cls=None,
            sig_generator_cls=None,
            **kwargs):
        """Constructs a new instance of LLHRatioAnalysis.

        Parameters
        ----------
        shg_mgr : instance of SourceHypoGroupManager
            The instance of SourceHypoGroupManager, which defines the groups of
            source hypotheses, their flux model, and their detector signal
            yield implementation method.
        pmm : instance of ParameterModelMapper
            The ParameterModelMapper instance managing the global set of
            parameters and their relation to individual models, e.g. sources.
        test_statistic : TestStatistic instance
            The TestStatistic instance that defines the test statistic function
            of the analysis.
        bkg_generator_cls : class of BackgroundGeneratorBase | None
            The background generator class used to create the background
            generator instance.
            If set to ``None``, the
            :class:`skyllh.core.background_generator.BackgroundGenerator` class
            is used.
        sig_generator_cls : class of SignalGenerator | None
            The signal generator class used to create the signal generator
            instance.
            If set to None, the
            :class:`~skyllh.core.signal_generator.MultiDatasetSignalGenerator`
            class is used.
        """
        super().__init__(
            shg_mgr=shg_mgr,
            pmm=pmm,
            test_statistic=test_statistic,
            bkg_generator_cls=bkg_generator_cls,
            sig_generator_cls=sig_generator_cls,
            **kwargs)

        # Define the member variable for the list of PDFRatio instances, one for
        # each dataset.
        self._pdfratio_list = []

        self._llhratio = None

    @property
    def llhratio(self):
        """The log-likelihood ratio function instance. It is None, if it has
        not been constructed yet.
        """
        if self._llhratio is None:
            raise RuntimeError(
                'The log-likelihood ratio function is not defined yet. '
                'Call the "construct_llhratio" method first!')
        return self._llhratio

    @llhratio.setter
    def llhratio(self, obj):
        if not isinstance(obj, LLHRatio):
            raise TypeError(
                'The llhratio property must be an instance of LLHRatio! '
                f'Its current type is {classname(obj)}.')
        self._llhratio = obj

    @abc.abstractmethod
    def construct_llhratio(
            self,
            minimizer,
            ppbar=None):
        """This method is supposed to construct the LLH ratio function.

        Returns
        -------
        llhratio : instance of LLHRatio
            The instance of LLHRatio that implements the
            log-likelihood-ratio function of this LLH ratio analysis.
        """
        pass

    def add_dataset(
            self,
            dataset,
            data,
            pdfratio,
            tdm=None,
            event_selection_method=None,
            bkg_generator=None,
            sig_generator=None):
        """Adds a dataset with its PDF ratio instances to the analysis.

        Parameters
        ----------
        dataset : instance of Dataset
            The instance of Dataset that should get added.
        data : instance of DatasetData
            The instance of DatasetData holding the original (prepared) data of
            the dataset.
        pdfratio : instance of PDFRatio
            The instance of PDFRatio for the to-be-added data set.
        tdm : instance of TrialDataManager | None
            The TrialDataManager instance that manages the trial data and
            additional data fields for this data set.
        event_selection_method : instance of EventSelectionMethod | None
            The instance of EventSelectionMethod to use to select only
            signal-like events from the trial data. All other events
            will be treated as pure background events. This reduces the amount
            of log-likelihood-ratio function evaluations. If set to None, all
            events will be evaluated.
        bkg_generator : instance of BackgroundGenerator | None
            The optional instance of BackgroundGenerator, which should be used
            to generate background events for this particular dataset.
        sig_generator : instance of SignalGenerator | None
            The optional instance of SignalGenerator, which should be used
            to generate signal events for this particular dataset.
        """
        super().add_dataset(
            dataset=dataset,
            data=data,
            tdm=tdm,
            event_selection_method=event_selection_method,
            bkg_generator=bkg_generator,
            sig_generator=sig_generator)

        if not isinstance(pdfratio, PDFRatio):
            raise TypeError(
                'The pdfratio argument must be an instance of PDFRatio! '
                f'Its current type is {classname(pdfratio)}')

        self._pdfratio_list.append(pdfratio)

    def change_shg_mgr(
            self,
            shg_mgr,
            update_detsigyield_service=True):
        """If the SourceHypoGroupManager instance changed, this method needs to
        be called to propagate the change to all components of the analysis.

        Parameters
        ----------
        shg_mgr : instance of SourceHypoGroupManager
            The new instance of SourceHypoGroupManager.
        update_detsigyield_service : bool
            The option whether to update SourceHypoGroupManager of
            `detsigyield_service` property. It can be set to false for runtime
            optimization when the detector signal yield does not change for the
            new source. Default is True.
        """
        if self._llhratio is None:
            raise RuntimeError(
                'The LLH ratio function has to be constructed '
                'before the `change_shg_mgr` method can be called!')

        super().change_shg_mgr(
            shg_mgr=shg_mgr,
            update_detsigyield_service=update_detsigyield_service)

        # Change the source hypo group manager of the LLH ratio function
        # instance.
        self._llhratio.change_shg_mgr(
            shg_mgr=shg_mgr)

    def initialize_trial(
            self,
            events_list,
            n_events_list=None,
            tl=None):
        """This method initializes the log-likelihood ratio
        function with a new set of given trial data. This is a low-level method.
        For convenient methods see the ``unblind`` and ``do_trial`` methods.

        Parameters
        ----------
        events_list : list of DataFieldRecordArray instances
            The list of DataFieldRecordArray instances holding the data events
            to use for the log-likelihood function evaluation. The data arrays
            for the datasets must be in the same order than the added datasets.
        n_events_list : list of int | None
            The list of the number of events of each data set. If set to None,
            the number of events is taken from the size of the given events
            arrays.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used for timing
            measurements.
        """
        if n_events_list is None:
            n_events_list = [None] * len(events_list)

        for (tdm, events, n_events, evt_sel_method) in zip(
                self._tdm_list,
                events_list,
                n_events_list,
                self._event_selection_method_list):

            # Initialize the trial data manager with the given raw events.
            tdm.initialize_trial(
                shg_mgr=self._shg_mgr,
                pmm=self._pmm,
                events=events,
                n_events=n_events,
                evt_sel_method=evt_sel_method,
                tl=tl)

        self._llhratio.initialize_for_new_trial(
            tl=tl)

    def unblind(
            self,
            minimizer_rss,
            tl=None):
        """Evaluates the unscrambled data, i.e. unblinds the data.

        Parameters
        ----------
        minimizer_rss : instance of RandomStateService
            The instance of RandomStateService that should be used by the
            minimizer to generate new random initial fit parameter values.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to time the
            maximization of the LLH ratio function.

        Returns
        -------
        TS : float
            The test-statistic value.
        global_params_dict : dict
            The dictionary holding the global parameter names and their
            best fit values. It includes fixed and floating parameters.
        status : dict
            The status dictionary with information about the performed
            minimization process of the negative of the log-likelihood ratio
            function.
        """
        events_list = [data.exp for data in self._data_list]
        self.initialize_trial(events_list)

        (log_lambda, fitparam_values, status) = self._llhratio.maximize(
            rss=minimizer_rss,
            tl=tl)

        TS = self.calculate_test_statistic(
            log_lambda=log_lambda,
            fitparam_values=fitparam_values)

        global_params_dict = self._pmm.create_global_params_dict(
            gflp_values=fitparam_values)

        return (TS, global_params_dict, status)

    def do_trial_with_given_pseudo_data(
            self,
            seed,
            mean_n_sig,
            n_sig,
            n_events_list,
            events_list,
            minimizer_rss,
            minimizer_status_dict=None,
            tl=None,
            mean_n_sig_0=None,
    ):
        """Performs an analysis trial on the given pseudo data.

        Parameters
        ----------
        seed : int
            The seed value, which was used to generate the pseudo data. It will
            be stored in the returned result array.
        mean_n_sig : float
            The mean number of signal events the pseudo data was generated with.
        n_sig : int
            The total number of actual signal events in the pseudo data.
        n_events_list : list of int
            The total number of events for each data set of the pseudo data.
        events_list : list of instance of DataFieldRecordArray
            The list of instance of DataFieldRecordArray containing the pseudo
            data events for each data sample. The number of events for each
            data sample can be less than the number of events given by
            ``n_events_list`` if an event selection method was already utilized
            when generating background events.
        minimizer_rss : instance of RandomStateService
            The instance of RandomStateService to use for generating random
            numbers for the minimizer, e.g. for new initial fit parameter
            values.
        minimizer_status_dict : dict | None
            If a dictionary is provided, it will be updated with the minimizer
            status dictionary.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to time
            individual tasks.
        mean_n_sig_0 : float | None
            The fixed mean number of signal events for the null-hypothesis,
            when using a ns-profile log-likelihood-ratio function.
            If set to None, this argument is interpreted as 0.

        Returns
        -------
        recarray : instance of numpy record ndarray
            The numpy record ndarray holding the result of the trial. It
            contains the following data fields:

            seed : int
                The seed value of the RandomStateService.
            mean_n_sig : float
                The mean number of signal events.
            n_sig : int
                The actual number of injected signal events.
            mean_n_sig_0 : float
                The fixed mean number of signal events for the null-hypothesis.
            ts : float
                The test-statistic value.
            [<global_param_name> : float ]
                Any additional parameters of the LLH ratio function.
        """
        if mean_n_sig_0 is None:
            mean_n_sig_0 = 0

        self._llhratio.mean_n_sig_0 = mean_n_sig_0

        with TaskTimer(tl, 'Initializing trial.'):
            self.initialize_trial(events_list, n_events_list)

        with TaskTimer(tl, 'Maximizing LLH ratio function.'):
            (log_lambda, fitparam_values, status) = self._llhratio.maximize(
                rss=minimizer_rss,
                tl=tl)
        if isinstance(minimizer_status_dict, dict):
            minimizer_status_dict.update(status)

        with TaskTimer(tl, 'Calculating test statistic.'):
            ts = self.calculate_test_statistic(
                log_lambda=log_lambda,
                fitparam_values=fitparam_values)

        # Get the dictionary holding all floating and fixed parameter names
        # and values.
        global_params_dict = self._pmm.create_global_params_dict(
            gflp_values=fitparam_values)

        # Create the structured array data type for the result array.
        recarray_dtype = [
            ('seed', np.int64),
            ('mean_n_sig', np.float64),
            ('n_sig', np.int64),
            ('mean_n_sig_0', np.float64),
            ('ts', np.float64)
        ] + [
            (param_name, np.float64)
            for param_name in global_params_dict.keys()
        ]
        recarray = np.empty((1,), dtype=recarray_dtype)
        recarray['seed'] = seed
        recarray['mean_n_sig'] = mean_n_sig
        recarray['n_sig'] = n_sig
        recarray['mean_n_sig_0'] = mean_n_sig_0
        recarray['ts'] = ts
        for (param_name, param_value) in global_params_dict.items():
            recarray[param_name] = param_value

        return recarray


class SingleSourceMultiDatasetLLHRatioAnalysis(
        LLHRatioAnalysis):
    """This is an analysis class that implements a log-likelihood ratio analysis
    for multiple datasets assuming a single source.
    It is a special case of the multi-source analysis.

    For more information how to construct and run the analysis see the
    documentation of the :class:`~skyllh.core.analysis.LLHRatioAnalysis` class.
    """
    def __init__(
            self,
            shg_mgr,
            pmm,
            test_statistic,
            bkg_generator_cls=None,
            sig_generator_cls=None,
            **kwargs):
        """Creates a new time-integrated point-like source analysis assuming a
        single source.

        Parameters
        ----------
        shg_mgr : instance of SourceHypoGroupManager
            The instance of SourceHypoGroupManager, which defines the groups of
            source hypotheses, their flux model, and their detector signal
            efficiency implementation method.
        pmm : instance of ParameterModelMapper
            The ParameterModelMapper instance managing the global set of
            parameters and their relation to individual models, e.g. sources.
        test_statistic : TestStatistic instance
            The TestStatistic instance that defines the test statistic function
            of the analysis.
        bkg_generator_cls : class of BackgroundGeneratorBase | None
            The background generator class used to create the background
            generator instance.
            If set to ``None``, the
            :class:`skyllh.core.background_generator.BackgroundGenerator` class
            is used.
        sig_generator_cls : SignalGenerator class | None
            The signal generator class that should be used to create the signal
            generator instance for multiple datasets. If set to None, the
            :class:`~skyllh.core.signal_generator.MultiDatasetSignalGenerator`
            class is used.
        """
        super().__init__(
            shg_mgr=shg_mgr,
            pmm=pmm,
            test_statistic=test_statistic,
            bkg_generator_cls=bkg_generator_cls,
            sig_generator_cls=sig_generator_cls,
            **kwargs)

    def construct_llhratio(
            self,
            minimizer,
            ppbar=None):
        """Constructs the log-likelihood (LLH) ratio function of the analysis.
        This setups all the necessary analysis objects like detector signal
        yields and dataset signal weights, constructs the log-likelihood ratio
        functions for each dataset and the final composite LLH ratio function.

        Parameters
        ----------
        minimizer : instance of Minimizer
            The instance of Minimizer that should be used to minimize the
            negative of the log-likelihood ratio function.
        ppbar : instance of ProgressBar | None
            The instance of ProgressBar of the optional parent progress bar.

        Returns
        -------
        llhratio : instance of MultiDatasetTCLLHRatio
            The instance of MultiDatasetTCLLHRatio that implements the
            log-likelihood-ratio function of the analysis.
        """
        # Create the list of log-likelihood ratio functions, one for each
        # dataset.
        llhratio_list = [
            ZeroSigH0SingleDatasetTCLLHRatio(
                cfg=self._cfg,
                pmm=self._pmm,
                minimizer=minimizer,
                shg_mgr=self._shg_mgr,
                tdm=tdm,
                pdfratio=pdfratio
            )
            for (tdm, pdfratio) in zip(self._tdm_list, self._pdfratio_list)
        ]

        # Create the final multi-dataset log-likelihood ratio function.
        llhratio = MultiDatasetTCLLHRatio(
            cfg=self._cfg,
            pmm=self._pmm,
            minimizer=minimizer,
            src_detsigyield_weights_service=self.src_detsigyield_weights_service,
            ds_sig_weight_factors_service=self.ds_sig_weight_factors_service,
            llhratio_list=llhratio_list)

        return llhratio

    def change_source(
            self,
            source,
            update_detsigyield_service=True):
        """Changes the source of the analysis to the given source. It makes the
        necessary changes to all the objects of the analysis.

        Parameters
        ----------
        source : instance of SourceModel
            The instance of SourceModel describing the new source.
        update_detsigyield_service : bool
            The option whether to update SourceHypoGroupManager of
            `detsigyield_service` property. It can be set to false for runtime
            optimization when the detector signal yield does not change for the
            new source. Default is True.
        """
        if not isinstance(source, SourceModel):
            raise TypeError(
                'The source argument must be an instance of SourceModel! '
                f'Its current type is {classname(source)}.')

        # Change the source in the SourceHypoGroupManager instance.
        # Because this is a single source analysis, there can only be one source
        # hypothesis group defined.
        self._shg_mgr.shg_list[0].source_list[0] = source

        self.change_shg_mgr(
            shg_mgr=self._shg_mgr,
            update_detsigyield_service=update_detsigyield_service)

    def calculate_fluxmodel_scaling_factor(
            self,
            mean_ns,
            fitparam_values):
        """Calculates the factor the source's fluxmodel has to be scaled
        in order to obtain the given mean number of signal events in the
        detector.

        Parameters
        ----------
        mean_ns : float
            The mean number of signal events in the detector for which the
            scaling factor is calculated.
        fitparam_values : instance of numpy ndarray
            The (N_fitparam,)-shaped 1D ndarray holding the values of the global
            fit parameters, that should be used for the flux calculation.
            The order of the values must match the order the fit parameters were
            defined in the parameter model mapper.

        Returns
        -------
        factor : float
            The factor the source's fluxmodel needs to be scaled in order to
            obtain the given mean number of signal events in the detector.
        """
        src_params_recarray =\
            self._pmm.create_src_params_recarray(
                gflp_values=fitparam_values)

        # Calculate the detector signal yield, i.e. the mean number of signal
        # events in the detector, for the given reference flux model.
        mean_ns_ref = 0

        detsigyields = self.detsigyield_service.arr[:, 0]
        for (j, detsigyield) in enumerate(detsigyields):
            src_recarray =\
                self.src_detsigyield_weights_service.src_recarray_list_list[j][0]
            (Yj, Yj_grads) = detsigyield(
                src_recarray=src_recarray,
                src_params_recarray=src_params_recarray)
            mean_ns_ref += Yj[0]

        factor = mean_ns / mean_ns_ref

        return factor


class MultiSourceMultiDatasetLLHRatioAnalysis(
        LLHRatioAnalysis):
    """This is an analysis class that implements a log-likelihood ratio analysis
    for multiple datasets assuming a multiple sources.

    For more information how to construct and run the analysis see the
    documentation of the :class:`~skyllh.core.analysis.LLHRatioAnalysis` class.
    """
    def __init__(
            self,
            shg_mgr,
            pmm,
            test_statistic,
            bkg_generator_cls=None,
            sig_generator_cls=None,
            **kwargs):
        """Constructs a new instance of MultiDatasetLLHRatioAnalysis.

        Parameters
        ----------
        shg_mgr : instance of SourceHypoGroupManager
            The instance of SourceHypoGroupManager, which defines the groups of
            source hypotheses, their flux model, and their detector signal
            yield implementation method.
        pmm : instance of ParameterModelMapper
            The ParameterModelMapper instance managing the global set of
            parameters and their relation to individual models, e.g. sources.
        test_statistic : TestStatistic instance
            The TestStatistic instance that defines the test statistic function
            of the analysis.
        bkg_generator_cls : class of BackgroundGeneratorBase | None
            The background generator class used to create the background
            generator instance.
            If set to ``None``, the
            :class:`skyllh.core.background_generator.BackgroundGenerator` class
            is used.
        sig_generator_cls : subclass of SignalGenerator| None
            The signal generator class that should be used to create the signal
            generator instance handling multiple datasets.
            If set to None, the
            :class:`~skyllh.core.signal_generator.MultiDatasetSignalGenerator`
            class is used.
        """
        super().__init__(
            shg_mgr=shg_mgr,
            pmm=pmm,
            test_statistic=test_statistic,
            bkg_generator_cls=bkg_generator_cls,
            sig_generator_cls=sig_generator_cls,
            **kwargs)

    def construct_llhratio(
            self,
            minimizer,
            ppbar=None):
        """Constructs the log-likelihood (LLH) ratio function of the analysis.
        This setups all the necessary analysis objects like detector signal
        yields and dataset signal weights, constructs the log-likelihood ratio
        functions for each dataset and the final composite LLH ratio function.

        Parameters
        ----------
        minimizer : instance of Minimizer
            The instance of Minimizer that should be used to minimize the
            negative of the log-likelihood ratio function.
        ppbar : instance of ProgressBar | None
            The instance of ProgressBar of the optional parent progress bar.

        Returns
        -------
        llhratio : instance of MultiDatasetTCLLHRatio
            The instance of MultiDatasetTCLLHRatio that implements the
            log-likelihood-ratio function of the analysis.
        """
        # Create the list of log-likelihood ratio functions, one for each
        # dataset.
        llhratio_list = [
            ZeroSigH0SingleDatasetTCLLHRatio(
                cfg=self._cfg,
                pmm=self._pmm,
                minimizer=minimizer,
                shg_mgr=self._shg_mgr,
                tdm=tdm,
                pdfratio=SourceWeightedPDFRatio(
                    cfg=self._cfg,
                    dataset_idx=dataset_idx,
                    src_detsigyield_weights_service=self.src_detsigyield_weights_service,
                    pdfratio=pdfratio)
            )
            for (dataset_idx, (tdm, pdfratio)) in enumerate(
                zip(self._tdm_list, self._pdfratio_list))
        ]

        # Create the final multi-dataset log-likelihood ratio function.
        llhratio = MultiDatasetTCLLHRatio(
            cfg=self._cfg,
            pmm=self._pmm,
            minimizer=minimizer,
            src_detsigyield_weights_service=self.src_detsigyield_weights_service,
            ds_sig_weight_factors_service=self.ds_sig_weight_factors_service,
            llhratio_list=llhratio_list)

        return llhratio

    def calculate_fluxmodel_scaling_factors(
            self,
            mean_ns,
            fitparam_values):
        """Calculates the factors the source's fluxmodel has to be scaled
        in order to obtain the given mean number of signal events in the
        detector.

        Parameters
        ----------
        mean_ns : float
            The mean number of signal events in the detector for which the
            scaling factors will be calculated.
        fitparam_values : instance of numpy ndarray
            The (N_fitparam,)-shaped 1D ndarray holding the values of the global
            fit parameters, which should be used for the flux calculation.
            The order of the values must match the order the fit parameters were
            defined in the parameter model mapper.

        Returns
        -------
        factors : instance of numpy ndarray
            The (N_sources,)-shaped numpy ndarray of float holding the factors
            the flux models of the sources need to be scaled in order to obtain
            the given mean number of signal events in the detector.
        """
        src_params_recarray =\
            self._pmm.create_src_params_recarray(
                gflp_values=fitparam_values)

        # Calculate the detector signal yield, i.e. the mean number of signal
        # events in the detector, for the given reference flux model.
        mean_ns_ref = np.zeros((self._shg_mgr.n_sources,), dtype=np.float64)

        for (g, shg) in enumerate(self._shg_mgr.shg_list):
            shg_src_mask = self._shg_mgr.get_src_mask_of_shg(shg_idx=g)

            detsigyields = self.detsigyield_service.arr[:, g]
            for (j, detsigyield) in enumerate(detsigyields):
                src_recarray =\
                    self.src_detsigyield_weights_service.src_recarray_list_list[j][g]
                (Yj, Yj_grads) = detsigyield(
                    src_recarray=src_recarray,
                    src_params_recarray=src_params_recarray)
                mean_ns_ref[shg_src_mask] += Yj

        factors = mean_ns / mean_ns_ref

        return factors



# -*- coding: utf-8 -*-

import numpy as np

from skyllh.core.livetime import (
    Livetime,
)
from skyllh.core.storage import (
    create_FileLoader,
)
from skyllh.i3.dataset import (
    I3Dataset,
)


class I3Livetime(
        Livetime):
    """The I3Livetime class provides the functionality to load a Livetime object
    from a good-run-list data file.
    """

    @classmethod
    def from_grl_data(cls, grl_data):
        """Creates an I3LiveTime instance from the given good-run-list (GRL)
        data.

        Parameters
        ----------
        grl_data : instance of numpy structured ndarray.
            The numpy structured ndarray of length N_runs holding the start end
            end times of the good runs. The following fields need to exist:

            start : float
                The MJD of the run start.
            end : float
                The MJD of the run stop.

        Returns
        -------
        livetime : instance of I3Livetime
            The created instance of I3Livetime for the provided GRL data.
        """
        uptime_mjd_intervals_arr = np.hstack((
            grl_data['start'].reshape((len(grl_data), 1)),
            grl_data['stop'].reshape((len(grl_data), 1))
        ))

        livetime = cls(
            uptime_mjd_intervals_arr=uptime_mjd_intervals_arr)

        return livetime

    @staticmethod
    def from_grl_files(
            pathfilenames):
        """Loads an I3Livetime instance from the given good-run-list (GRL) data
        file. The data file needs to contain the following data fields:

            start : float
                The MJD of the run start.
            stop : float
                The MJD of the run stop.

        Parameters
        ----------
        pathfilenames : str | list of str
            The list of fully qualified file names of the GRL data files.

        Returns
        -------
        livetime : instance of I3Livetime
            The created instance of I3Livetime for the provided GRL data.
        """
        grl_data = create_FileLoader(pathfilenames).load_data()

        uptime_mjd_intervals_arr = np.hstack((
            grl_data['start'].reshape((len(grl_data), 1)),
            grl_data['stop'].reshape((len(grl_data), 1))
        ))

        livetime = I3Livetime(
            uptime_mjd_intervals_arr=uptime_mjd_intervals_arr)

        return livetime

    @staticmethod
    def from_I3Dataset(ds):
        """Loads an I3Livetime instance from a given I3Dataset instance, which
        must have a good-run-list (GRL) files defined.

        Parameters
        ----------
        ds : I3Dataset instance
            The instance of I3Dataset which defined the good-run-list (GRL)
            files for the dataset.

        Returns
        -------
        livetime : instance of I3Livetime
            The created instance of I3Livetime for the GRL data from the
            provided dataset.
        """
        if not isinstance(ds, I3Dataset):
            raise TypeError(
                'The ds argument must be an instance of I3Dataset!')
        if len(ds.grl_pathfilename_list) == 0:
            raise ValueError(
                'No GRL files have been defined for the given dataset!')

        livetime = I3Livetime.from_grl_files(
            pathfilenames=ds.grl_pathfilename_list)

        return livetime

    def __init__(
            self,
            *args,
            **kwargs):
        """Creates a new instance of I3Livetime.
        """
        super().__init__(
            *args,
            **kwargs)
